import logging
import os

from spinn_front_end_common.utility_models.\
    chip_power_monitor_machine_vertex import \
    ChipPowerMonitorMachineVertex
from spinn_front_end_common.utilities import exceptions, helpful_functions

logger = logging.getLogger(__name__)


class EnergyReport(object):

    # given from indar measurements
    JULES_PER_MILLISECOND_PER_FPGA = 0.000584635

    # stated in papers (SpiNNaker: A 1-W 18 core system-on-Chip for
    # Massively-Parallel Neural Network Simulation)
    JULES_PER_SPIKE = 0.000000000800

    # stated in papers (SpiNNaker: A 1-W 18 core system-on-Chip for
    # Massively-Parallel Neural Network Simulation)
    JULES_PER_MILLISECOND_PER_IDLE_CHIP = 0.000360

    # stated in papers (SpiNNaker: A 1-W 18 core system-on-Chip for
    # Massively-Parallel Neural Network Simulation)
    JULES_PER_MILLISECOND_PER_CHIP_ACTIVE_OVERHEAD = \
        0.001 - JULES_PER_MILLISECOND_PER_IDLE_CHIP

    # converter between jules to Killiwatt hours
    JULES_TO_KILLIWATT_HOURS = 3600000

    # measured from the real power meter and timing between
    #  the photos for a days powered off
    JULES_PER_MILLISECOND_FOR_FRAME_IDLE_COST = 0.117

    # measured from the loading of the column and extrapolated
    JULES_PER_MILLISECOND_PER_FRAME_ACTIVE_COST = 0.154163558

    # measured from the real power meter and timing between the photos
    # for a day powered off
    JULES_PER_MILLISECOND_FOR_BOXED_48_CHIP_FRAME_IDLE_COST = 0.0045833333

    # TODO needs filling in
    JULES_PER_MILLISECOND_PER_UNBOXED_48_CHIP_FRAME_IDLE_COST = 0.01666667

    # TODO verify this is correct when doing multiboard comms
    N_MONITORS_ACTIVE_DURING_COMMS = 2

    # energy report file name
    ENERGY_DETAILED_FILENAME = "Detailed_energy_report.rpt"
    ENERGY_SUMMARY_FILENAME = "energy_summary_report.rpt"

    def __call__(
            self, placements, machine, report_default_directory, version,
            spalloc_server, remote_spinnaker_url, time_scale_factor,
            machine_time_step, pacman_provenance, router_provenance,
            machine_graph, runtime, buffer_manager, mapping_time, load_time,
            execute_time, dsg_time, extraction_time,
            machine_allocation_controller=None):
        """ main call

        :param placements: the placements
        :param machine: the machine
        :param report_default_directory: location for reports
        :param version: version of machine
        :param spalloc_server: spalloc server ip
        :param remote_spinnaker_url: remote spinnaker url
        :param time_scale_factor: the time scale factor
        :param machine_time_step: the machine time step
        :param pacman_provenance: the pacman provenance
        :param router_provenance: the router provenance
        :param machine_graph: the machine graph
        :param buffer_manager: the buffer manager
        :param mapping_time: time taken by the mapping process
        :param load_time: the time taken by the load process
        :param execute_time: the time taken by the execute time process
        :param dsg_time: the time taken by the dsg time
        :param extraction_time: the time taken by data extraction time
        :param machine_allocation_controller: the machine controller for
        spalloc
        :rtype: None
        """

        if buffer_manager is None:
            logger.info("Skipping Energy report as no buffer_manager set")
            return

        # detailed report path
        detailed_report = os.path.join(
            report_default_directory, self.ENERGY_DETAILED_FILENAME)

        # summary report path
        summary_report = os.path.join(
            report_default_directory, self.ENERGY_SUMMARY_FILENAME)

        # overall time taken up
        total_time = (
            execute_time + load_time + extraction_time + dsg_time +
            mapping_time)

        # total time the machine was booted
        total_booted_time = execute_time + load_time + extraction_time

        # figure runtime in milliseconds with time scale factor
        runtime_total_milliseconds = runtime * time_scale_factor

        # create detailed report
        with open(detailed_report, "w") as output:
            active_chip_cost, fpga_cost_total, fpga_cost_runtime, \
                packet_cost, mapping_cost, load_time_cost, \
                data_extraction_cost, dsg_cost, router_cooling_runtime_cost = \
                self._write_detailed_report(
                    placements, machine, version, spalloc_server,
                    remote_spinnaker_url,
                    pacman_provenance, router_provenance, dsg_time,
                    buffer_manager, output, load_time, mapping_time,
                    total_booted_time, machine_allocation_controller,
                    runtime_total_milliseconds)

        # create summary report
        with open(summary_report, "w") as output:
            self._write_summary_report(
                active_chip_cost, fpga_cost_total, fpga_cost_runtime,
                packet_cost, mapping_cost, load_time_cost,
                data_extraction_cost, runtime_total_milliseconds, output,
                mapping_time, load_time, dsg_time, dsg_cost,
                extraction_time, total_time, total_booted_time,
                router_cooling_runtime_cost)

    @staticmethod
    def _write_summary_report(
            active_chip_cost, fpga_cost_total, fpga_cost_runtime, packet_cost,
            mapping_cost, load_time_cost, data_extraction_cost,
            runtime_total_milliseconds, output, mapping_time, load_time,
            dsg_time, dsg_cost, extraction_time, total_time,
            total_booted_time, router_cooling_runtime_cost):
        """ write summary file

        :param active_chip_cost: active chip cost
        :param fpga_cost_total: fpga cost over all booted time
        :param fpga_cost_runtime: fpga cost during runtime
        :param mapping_time: the time taken by the mapping process in ms
        :param mapping_cost: the energy used by the mapping process
        :param packet_cost: packet cost
        :param load_time_cost: load time cost
        :param data_extraction_cost: data extraction cost
        :param runtime_total_milliseconds: runtime with time scale factor
            taken into account
        :param output: file writer
        :rtype: None
        """

        # total the energy costs
        total_jules = (
            active_chip_cost + fpga_cost_total + packet_cost + mapping_cost +
            load_time_cost + data_extraction_cost + dsg_cost +
            fpga_cost_runtime + router_cooling_runtime_cost)

        # deduce wattage from the runtime
        total_watts = total_jules / (total_time / 1000)

        # figure total killawatt hour
        killawatt_hour = total_jules / EnergyReport.JULES_TO_KILLIWATT_HOURS

        # write summary data
        output.write(
            "Summary energy file\n\n"
            "Energy used by chips during runtime is {} Joules over {} "
            "milliseconds\n"
            "Energy used by FPGAs is {} Joules over the entire time the "
            "machine was booted is {} milliseconds \n"
            "Energy used by FPGAs is {} Joules over the runtime period is "
            "{} milliseconds \n"
            "Energy used by outside router / cooling during the runtime "
            "period is {} Joules\n"
            "Energy used by packet transmissions is {} Joules over {} "
            "milliseconds\n"
            "Energy used during the mapping process is {} Joules over {} "
            "milliseconds\n"
            "Energy used by the data generation process is {} Joules over {} "
            "milliseconds\n"
            "Energy used during the loading process is {} Joules over {} "
            "milliseconds\n"
            "Energy used during the data extraction process is {} Jules over "
            "{} milliseconds\n"
            "Total energy used by the simulation over {} milliseconds is: \n"
            "     {} Joules or \n"
            "     {} estimated average Watts or \n"
            "     {} kWh\n".format(
                active_chip_cost, runtime_total_milliseconds, fpga_cost_total,
                total_booted_time, fpga_cost_runtime,
                runtime_total_milliseconds, router_cooling_runtime_cost,
                packet_cost, total_time, mapping_cost, mapping_time, dsg_cost,
                dsg_time, load_time_cost, load_time, data_extraction_cost,
                extraction_time, total_time, total_jules, total_watts,
                killawatt_hour))

    def _write_detailed_report(
            self, placements, machine, version, spalloc_server,
            remote_spinnaker_url,
            pacman_provenance, router_provenance, dsg_time,
            buffer_manager, output, load_time, mapping_time,
            total_booted_time, machine_allocation_controller,
            runtime_total_milliseconds):
        """ write detailed report and calculate costs

        :param placements: placements
        :param machine: machine rep
        :param version: machine version
        :param spalloc_server: spalloc server
        :param remote_spinnaker_url: remote spinnaker url
        :param pacman_provenance: provenance generated by pacman
        :param router_provenance: provenance generated by the router
        :param buffer_manager: buffer manager
        :param output: file writer
        :param total_booted_time: time in milliseconds where machine is booted
        :param runtime_total_milliseconds: total runtime with time scale
        factor taken into account
        :return: machine_active_cost, machine_idle_chips_cost, \
            fpga_cost, packet_cost, load_time_cost, extraction_time_cost
        """

        # write warning about accuracy etc
        self._write_warning(output)

        # figure active chips
        active_chips = set()
        for placement in placements:
            if not isinstance(placement.vertex, ChipPowerMonitorMachineVertex):
                active_chips.add(machine.get_chip_at(placement.x, placement.y))

        # figure out packet cost
        packet_cost = self._router_packet_cost(router_provenance, output)

        # figure FPGA cost over all booted and during runtime cost
        fpga_cost_total, fpga_cost_runtime = self._calculate_fpga_cost(
            machine, version, spalloc_server, remote_spinnaker_url,
            total_booted_time, output, runtime_total_milliseconds)

        # figure how many frames are using, as this is a constant cost of
        # routers, cooling etc
        n_frames = self._calculate_n_frames(
            machine, machine_allocation_controller)

        # figure load time cost
        load_time_cost = self._calculate_load_time_cost(
            pacman_provenance, machine, output, load_time, active_chips,
            n_frames)

        # figure the down time idle cost for mapping
        mapping_cost = self._calculate_power_down_cost(
            mapping_time, machine, machine_allocation_controller, version,
            n_frames)

        # figure the down time idle cost for dsg
        dsg_cost = self._calculate_power_down_cost(
            dsg_time, machine, machine_allocation_controller, version,
            n_frames)

        # figure extraction time cost
        extraction_time_cost = \
            self._calculate_data_extraction_time_cost(
                pacman_provenance, machine, output, active_chips, n_frames)

        # figure out active chips idle time
        machine_active_cost = 0.0
        for chip in active_chips:
            machine_active_cost += self._calculate_chips_active_cost(
                chip, placements, buffer_manager, output,
                runtime_total_milliseconds)

        # figure out router idle cost during runtime
        router_cooling_runtime_cost = (
            runtime_total_milliseconds * n_frames *
            self.JULES_PER_MILLISECOND_FOR_FRAME_IDLE_COST)

        # return all magic values
        return machine_active_cost, fpga_cost_total, fpga_cost_runtime, \
            packet_cost, mapping_cost, load_time_cost, extraction_time_cost, \
            dsg_cost, router_cooling_runtime_cost

    def _write_warning(self, output):
        """ writes the warning about this being only an estimate

        :param output: the writer
        :rtype: None
        """

        output.write(
            "This report is based off energy estimates for individual "
            "components of the SpiNNaker machine. It is not meant to be "
            "completely accurate. But does use provenance data gathered from"
            " the machine to estimate the energy usage and therefore should "
            "be within the ball park.\n\n\n")
        output.write(
            "The energy components we use are as follows: \n\n"
            "The energy usage for a chip when all cores are 100% active for"
            " a millisecond is {} Jules\n"
            "The energy usage for a chip when all cores are not active for a "
            "millisecond is {} Jules\n"
            "The energy used by the machine for firing a packet is {} Jules\n"
            "The energy used by each active FPGA per millisecond is {} "
            "Jules.\n\n\n"
            .format(
                self.JULES_PER_MILLISECOND_PER_CHIP_ACTIVE_OVERHEAD,
                self.JULES_PER_MILLISECOND_PER_IDLE_CHIP,
                self.JULES_PER_SPIKE, self.JULES_PER_MILLISECOND_PER_FPGA))

    def _calculate_fpga_cost(
            self, machine, version, spalloc_server, remote_spinnaker_url,
            total_runtime, output, runtime_total_milliseconds):
        """ fpga cost calculation

        :param machine: machine rep
        :param version: machine version
        :param spalloc_server: spalloc server ip
        :param remote_spinnaker_url: remote spinnaker
        :param total_runtime: runtime:
        :param output: the file writer:
        :param runtime_total_milliseconds:
        :return: power usage of fpgas
        """

        # if not spalloc, then could be any type of board
        if spalloc_server is None and remote_spinnaker_url is None:

            # if a spinn2 or spinn3 (4 chip boards) then they have no fpgas
            if int(version) == 2 or int(version) == 3:
                output.write(
                    "A Spinn {} board does not contain any FPGA's, and so "
                    "its energy cost is 0 \n".format(version))
                return 0, 0

            # if the spinn4 or spinn5 board, need to verify if wrap arounds
            # are there, if not then assume fppga's are turned off.
            elif int(version) == 4 or int(version) == 5:

                # how many fpgas are active
                n_operational_fpgas = self._board_n_operational_fpgas(
                    machine, machine.ethernet_connected_chips[0])

                # active fpgas
                if n_operational_fpgas > 0:
                    return self._print_out_fpga_cost(
                        total_runtime, n_operational_fpgas, output, version,
                        runtime_total_milliseconds)
                else:  # no active fpgas
                    output.write(
                        "The FPGA's on the Spinn {} board are turned off and "
                        "therefore the energy used by the FPGA is 0\n".format(
                            version))
                    return 0, 0
            else:  # no idea where we are
                raise exceptions.ConfigurationException(
                    "Do not know what the FPGA setup is for this version of "
                    "SpiNNaker machine.")
        else:  # spalloc machine, need to check each board
            total_fpgas = 0
            for ethernet_connected_chip in machine.ethernet_connected_chips:
                total_fpgas += self._board_n_operational_fpgas(
                    machine, ethernet_connected_chip)
            return self._print_out_fpga_cost(
                total_runtime, total_fpgas, output, version,
                runtime_total_milliseconds)

    def _print_out_fpga_cost(
            self, total_runtime, n_operational_fpgas, output, version,
            runtime_total_milliseconds):
        """ prints out to file and returns cost

        :param total_runtime: all runtime
        :param n_operational_fpgas: n operational fpgas
        :param output: file writer
        :param version: machine version
        :param runtime_total_milliseconds: runtime in milliseconds
        :return: power usage
        """
        power_usage_total = (
            total_runtime * self.JULES_PER_MILLISECOND_PER_FPGA *
            n_operational_fpgas)
        power_usage_runtime = (
            runtime_total_milliseconds * self.JULES_PER_MILLISECOND_PER_FPGA *
            n_operational_fpgas)

        # print out as needed for spalloc and none spalloc versions
        if version is None:
            output.write(
                "{} FPGA's on the Spalloced boards are turned on and "
                "therefore the energy used by the FPGA during the entire time "
                "the machine was booted (which was {} ms) is {}. "
                "The usage during execution was {}".format(
                    n_operational_fpgas, total_runtime,
                    power_usage_total, power_usage_runtime))
        else:
            output.write(
                "{} FPGA's on the Spinn {} board are turned on and "
                "therefore the energy used by the FPGA during the entire time "
                "the machine was booted (which was {} ms) is {}. "
                "The usage during execution was {}".format(
                    n_operational_fpgas, version, total_runtime,
                    power_usage_total, power_usage_runtime))
        return power_usage_total, power_usage_runtime

    def _board_n_operational_fpgas(self, machine, ethernet_connected_chip):
        """ figures fpgas on

        :param machine: spinnaker machine
        :param ethernet_connected_chip: the ethernet chip to look from
        :return: number of fpgas on, on this board
        """

        # positions to check for active links
        left_additions = [[0, 0], [0, 1], [0, 2], [0, 3], [0, 4]]
        right_additions = [[7, 3], [7, 4], [7, 5], [7, 6], [7, 7]]
        top_additions = [[4, 7], [5, 7], [6, 7], [7, 7]]
        bottom_additions = [[0, 0], [1, 0], [2, 0], [3, 0], [4, 0]]
        top_left_additions = [[0, 3], [1, 4], [2, 5], [3, 6], [4, 7]]
        bottom_right_additions = [[0, 4], [1, 5], [2, 6], [3, 7]]

        machine_max_x = machine.max_chip_x
        machine_max_y = machine.max_chip_y

        ethernet_chip_x = ethernet_connected_chip.x
        ethernet_chip_y = ethernet_connected_chip.y

        # bottom left, bottom
        fpga_0 = self._deduce_fpga(
            [bottom_additions, bottom_right_additions], [[5, 4], [0, 5]],
            machine_max_x, machine_max_y, ethernet_chip_x, ethernet_chip_y,
            machine)
        # left, and top right
        fpga_1 = self._deduce_fpga(
            [left_additions, top_left_additions], [[3, 4], [3, 2]],
            machine_max_x, machine_max_y, ethernet_chip_x, ethernet_chip_y,
            machine)
        # top and right
        fpga_2 = self._deduce_fpga(
            [top_additions, right_additions], [[2, 1], [0, 1]],
            machine_max_x, machine_max_y, ethernet_chip_x, ethernet_chip_y,
            machine)
        return fpga_1 + fpga_0 + fpga_2

    @staticmethod
    def _deduce_fpga(
            shifts, overall_link_ids, machine_max_x, machine_max_y,
            ethernet_chip_x, ethernet_chip_y, machine):
        """ figure each fpga on or not

        :param shifts: shifts from ethernet to find a fpga edge
        :param overall_link_ids: which link ids to check
        :param machine_max_x: max machine x
        :param machine_max_y: max machine y
        :param ethernet_chip_x: ethernet chip x
        :param ethernet_chip_y: ethernet chip y
        :param machine: machine rep
        :return: 0 if not on, 1 if on
        """
        for shift_group, link_ids in zip(shifts, overall_link_ids):
            for shift in shift_group:
                new_x = (ethernet_chip_x + shift[0]) % (machine_max_x + 1)
                new_y = (ethernet_chip_y + shift[1]) % (machine_max_y + 1)
                chip = machine.get_chip_at(new_x, new_y)
                if chip is not None:
                    for link_id in link_ids:
                        link = chip.router.get_link(link_id)
                        if link is not None:
                            return 1
        return 0

    def _calculate_chips_active_cost(
            self, chip, placements, buffer_manager, output,
            runtime_total_milliseconds):
        """ figure chip active cost during sim

        :param chip: the chip to consider
        :param placements: placements
        :param buffer_manager: buffer manager
        :param output: file writer
        :return: energy cost
        """

        # locate chip power monitor
        chip_power_monitor = None

        # start at top, as more likely it was placed on the top
        processor_id = 18
        while chip_power_monitor is None:
            processor = chip.get_processor_with_id(processor_id)
            if (processor is not None and
                    placements.is_processor_occupied(
                        chip.x, chip.y, processor_id)):

                # check if vertex is a chip power monitor
                vertex = placements.get_vertex_on_processor(
                    chip.x, chip.y, processor_id)
                if isinstance(vertex, ChipPowerMonitorMachineVertex):
                    chip_power_monitor = vertex
            processor_id -= 1

        # get recordings from the chip power monitor
        recorded_measurements = chip_power_monitor.get_recorded_data(
            placement=placements.get_placement_of_vertex(chip_power_monitor),
            buffer_manager=buffer_manager)

        # deduce time in millsecodns per recording element
        time_for_recorded_sample = (
            chip_power_monitor.sampling_frequency *
            chip_power_monitor.n_samples_per_recording) / 1000
        cores_power_cost = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                            0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

        # accumulate costs
        for recorded_measurement in recorded_measurements:
            for core in range(0, 18):
                cores_power_cost[core] += (
                    recorded_measurement[core] * time_for_recorded_sample *
                    (self.JULES_PER_MILLISECOND_PER_CHIP_ACTIVE_OVERHEAD / 18))

        # detailed report print out
        for core in range(0, 18):
            output.write(
                "processor {}:{}:{} used {} Jules of energy by being active "
                "during the execution of the simulation\n".format(
                    chip.x, chip.y, core, cores_power_cost[core]))

        total_energy_cost = 0.0
        for core_power_usage in cores_power_cost:
            total_energy_cost += core_power_usage

        # TAKE INTO ACCOUNT IDLE COST
        idle_cost = (
            runtime_total_milliseconds *
            self.JULES_PER_MILLISECOND_PER_IDLE_CHIP)

        total_energy_cost += idle_cost

        output.write(
            "The machine used {} Jules of energy by being idle "
            "during the execution of the simulation".format(idle_cost))

        return total_energy_cost

    def _router_packet_cost(self, router_provenance, output):
        """ packet cost, includes MC, P2P, FR, NN packets

        :param router_provenance: the provenance gained from the router
        :param output: file writer
        :rtype: energy usage value
        """

        energy_cost = 0.0
        for element in router_provenance:

            # only get per chip counters, not summary counters.
            if (element.names[1] == "expected_routers" or
                    element.names[1] == "unexpected_routers"):

                # process multicast packets
                if (element.names[3] == "Local_Multicast_Packets" or
                        element.names[3] == "External_Multicast_Packets" or
                        element.names[3] == "Reinjected"):
                    energy_cost += float(element.value) * self.JULES_PER_SPIKE

                # process p2p packets
                elif (element.names[3] == "Local_P2P_Packets" or
                        element.names[3] == "External_P2P_Packets"):
                    energy_cost += \
                        float(element.value) * self.JULES_PER_SPIKE * 2

                # process NN packets
                elif (element.names[3] == "Local_NN_Packets" or
                        element.names[3] == "External_NN_Packets"):
                    energy_cost += float(element.value) * self.JULES_PER_SPIKE

                elif (element.names[3] == "Local_FR_Packets" or
                        element.names[3] == "External_FR_Packets"):
                    energy_cost += \
                        float(element.value) * self.JULES_PER_SPIKE * 2

        # detailed report print
        output.write("The packet cost is {} Jules\n".format(energy_cost))
        return energy_cost

    def _calculate_load_time_cost(
            self, pacman_provenance, machine, output, load_time,
            active_chips, n_frames):
        """ energy usage from the loading phase

        :param pacman_provenance: provenance items from the pacman set
        :param machine: machine rep
        :param output: file writer
        :param active_chips: the chips which have end user code in them
        :param load_time: the time of the entire load time phase in ms
        :return: load time energy value
        """

        # find time
        total_milliseconds = 0.0
        loading_algorithms = list()
        for element in pacman_provenance:
            if element.names[1] == "loading":
                loading_algorithms.append(element)

        for element in loading_algorithms:
            total_milliseconds += \
                helpful_functions.convert_time_diff_to_total_milliseconds(
                    element.value)

        # handle monitor core active cost
        # min between chips that are active and fixed monitor, as when 1
        # chip is used its one monitor, if more than 1 chip,
        # the ethernet connected chip and the monitor handling the read/write
        # this is checked by min
        energy_cost = (
            total_milliseconds *
            min(self.N_MONITORS_ACTIVE_DURING_COMMS, len(active_chips)) *
            (self.JULES_PER_MILLISECOND_PER_CHIP_ACTIVE_OVERHEAD / 18))

        # handle all idle cores
        energy_cost += self._calculate_idle_cost(total_milliseconds, machine)

        # handle time diff between load time and total laod phase of ASB
        diff_of_algorithms_and_boiler = load_time - total_milliseconds
        energy_cost += (
            diff_of_algorithms_and_boiler * (
                len(list(machine.chips)) *
                self.JULES_PER_MILLISECOND_PER_IDLE_CHIP))

        # handle active routers etc
        energy_cost_of_active_router = (
            load_time * n_frames *
            self.JULES_PER_MILLISECOND_PER_FRAME_ACTIVE_COST)

        # accumulate
        energy_cost += energy_cost_of_active_router

        # detailed report write
        output.write(
            "The amount of time used during the loading process is {} "
            "milliseconds.\n Assumed only 2 monitor cores is executing that "
            "this point. We also assume that there is a baseline active "
            "router/cooling component that is using {} Jules. "
            "Overall the energy usage is {} Jules \n".format(
                total_milliseconds, energy_cost_of_active_router, energy_cost))

        return energy_cost

    def _calculate_data_extraction_time_cost(
            self, pacman_provenance, machine, output, active_chips, n_frames):
        """ data extraction cost

        :param pacman_provenance: provenance items from the pacman set
        :param machine: machine rep
        :param output: file writer
        :param active_chips:
        :return: cost of data extraction
        """

        # find time
        total_milliseconds = 0.0
        extraction_algorithms = list()
        for element in pacman_provenance:
            if element.names[1] == "Execution":
                if not (element.names[2] ==
                        "run_time_of_FrontEndCommonApplicationRunner"):
                    extraction_algorithms.append(element)

        for element in extraction_algorithms:
            total_milliseconds += \
                helpful_functions.convert_time_diff_to_total_milliseconds(
                    element.value)

        # min between chips that are active and fixed monitor, as when 1
        # chip is used its one monitor, if more than 1 chip,
        # the ethernet connected chip and the monitor handling the read/write
        # this is checked by min
        energy_cost = (
            total_milliseconds *
            min(self.N_MONITORS_ACTIVE_DURING_COMMS, len(active_chips)) *
            (self.JULES_PER_MILLISECOND_PER_CHIP_ACTIVE_OVERHEAD / 18))

        # add idle chip cost
        energy_cost += self._calculate_idle_cost(total_milliseconds, machine)

        # handle active routers etc
        energy_cost_of_active_router = (
            total_milliseconds * n_frames *
            self.JULES_PER_MILLISECOND_PER_FRAME_ACTIVE_COST)
        energy_cost += energy_cost_of_active_router

        # detailed report
        output.write(
            "The amount of time used during the data extraction process is {} "
            "milliseconds.\n Assumed only 2 monitor cores is executing at "
            "this point. we also assume that there is a baseline active "
            "router/cooling component that is using {} Jules. so the overall "
            "energy usage is {} Jules \n".format(
                total_milliseconds, energy_cost_of_active_router, energy_cost))

        return energy_cost

    def _calculate_idle_cost(self, time, machine):
        """ calculate energy used by being idle

        :param machine: machine object
        :param time: time machine was idle
        :return: cost in jules
        """
        return time * machine.total_available_user_cores * (
                self.JULES_PER_MILLISECOND_PER_IDLE_CHIP / 18)

    def _calculate_power_down_cost(
            self, time, machine, machine_allocation_controller, version,
            n_frames):
        """ calculate power down costs

        :param time: time powered down
        :param n_frames: number of frames used by this machine
        :return: power in jules
        """

        # if spalloc or hbp
        if machine_allocation_controller is not None:
            return (time * n_frames *
                    self.JULES_PER_MILLISECOND_FOR_FRAME_IDLE_COST)
        # if 48 chip
        if version == 5 or version == 4:
            return (
                time *
                self.JULES_PER_MILLISECOND_FOR_BOXED_48_CHIP_FRAME_IDLE_COST)
        # if 4 chip
        if version == 3 or version == 2:
            return (len(list(machine.chips)) *
                    time * self.JULES_PER_MILLISECOND_PER_IDLE_CHIP)
        # boom
        raise exceptions.ConfigurationException("dont know what to do here")

    @staticmethod
    def _calculate_n_frames(machine, machine_allocation_controller):
        """ figures out how many frames are being used in this setup
        key of cabinate frame will be used to identify unique frame.

        :param machine: the machine object
        :param machine_allocation_controller: the spalloc job object
        :return: n frames
        """

        # if not spalloc, then could be any type of board, but unknown cooling
        if machine_allocation_controller is None:
            return 0

        # if using spalloc in some form
        cabinate_frame = set()
        for ethernet_connected_chip in machine.ethernet_connected_chips:
            cabinate, frame, _ = \
                machine_allocation_controller.where_is_machine(
                    chip_x=ethernet_connected_chip.x,
                    chip_y=ethernet_connected_chip.y
                )
            cabinate_frame.add((cabinate, frame))
        return len(list(cabinate_frame))



from spinn_front_end_common.interface.profiling.profile_data \
    import ProfileData

import logging
import struct

from spinn_front_end_common.utilities import helpful_functions

logger = logging.getLogger(__name__)

PROFILE_HEADER_SIZE_BYTES = 4
SIZE_OF_PROFILE_DATA_ENTRY_IN_BYTES = 8
BYTE_OFFSET_OF_PROFILE_DATA_IN_PROFILE_REGION = 4


def get_profile_region_size(n_samples):
    """ Get the size of the region of the profile data

    :param n_samples: number of different samples to record
    :return the size in bytes used by the profile region
    """
    return PROFILE_HEADER_SIZE_BYTES + (
        n_samples * SIZE_OF_PROFILE_DATA_ENTRY_IN_BYTES)


def reserve_profile_region(spec, region, n_samples):
    """ Reserves the profile region for recording the profile data

    :param spec: the dsg specification writer
    :param region: region id for the profile data
    :param n_samples: n elements being sampled
    :rtype: None

    """
    size = get_profile_region_size(n_samples)
    spec.reserve_memory_region(
        region=region, size=size, label="profilerRegion")


def write_profile_region_data(spec, region, n_samples):
    """ Writes the profile region data

    :param spec: the dsg specification writer
    :param region: region id for the profile data
    :param n_samples: n elements being sampled
    :rtype: None
    """
    spec.switch_write_focus(region)
    spec.write_value(n_samples)


def get_profiling_data(profile_region, tag_labels, txrx, placement):
    """ Utility function to get profile data from a profile region

    :param profile_region: dsg region to get profiling data out of sdram
    :param tag_labels: labels for the profiling data
    :param txrx:  transceiver code
    :param placement: placement
    :return: ProfileData
    """

    profile_data = ProfileData(tag_labels)

    profiling_region_base_address = \
        helpful_functions.locate_memory_region_for_placement(
            placement=placement, region=profile_region, transceiver=txrx)

    # Read the profiling data size
    words_written_data =\
        buffer(txrx.read_memory(
            placement.x, placement.y,
            profiling_region_base_address, 4))
    words_written = \
        struct.unpack_from("<I", words_written_data)[0]

    # Read the profiling data
    if words_written != 0:
        profiling_data = txrx.read_memory(
            placement.x, placement.y,
            profiling_region_base_address +
            BYTE_OFFSET_OF_PROFILE_DATA_IN_PROFILE_REGION,
            words_written * 4)
        profile_data.add_data(profiling_data)

    return profile_data



from .machine_data_specable_vertex import MachineDataSpecableVertex
from .needs_n_machine_time_steps import NeedsNMachineTimeSteps
from .provides_key_to_atom_mapping_impl import ProvidesKeyToAtomMappingImpl

__all__ = ["MachineDataSpecableVertex", "NeedsNMachineTimeSteps",
           "ProvidesKeyToAtomMappingImpl"]



from spinn_front_end_common.utilities.exceptions import ConfigurationException


class MultiCastCommand(object):
    """ A command to be sent to a vertex
    """

    def __init__(
            self, key, payload=None, time=None, repeat=0,
            delay_between_repeats=0):
        """

        :param key: The key of the command
        :type key: int
        :param payload: The payload of the command
        :type payload: int
        :param time: The time within the simulation at which to send the\
                    command, or None if this is not a timed command
        :type time: int
        :param repeat: The number of times that the command should be\
                    repeated after sending it once.  This could be used to\
                    ensure that the command is sent despite lost packets.\
                    Must be between 0 and 65535
        :type repeat: int
        :param delay_between_repeats: The amount of time in micro seconds to\
                    wait between sending repeats of the same command.\
                    Must be between 0 and 65535, and must be 0 if repeat is 0
        :type delay_between_repeats: int
        :raise SpynnakerException: If the repeat or delay are out of range
        """

        if repeat < 0 or repeat > 0xFFFF:
            raise ConfigurationException(
                "repeat must be between 0 and 65535")
        if delay_between_repeats < 0 or delay_between_repeats > 0xFFFF:
            raise ConfigurationException(
                "delay_between_repeats must be between 0 and 65535")
        if delay_between_repeats > 0 and repeat == 0:
            raise ConfigurationException(
                "If repeat is 0, delay_betweeen_repeats must be 0")

        self._time = time
        self._key = key

        self._payload = payload
        self._repeat = repeat
        self._delay_between_repeats = delay_between_repeats

    @property
    def time(self):
        return self._time

    @property
    def is_timed(self):
        return self._time is not None

    @property
    def key(self):
        return self._key

    @property
    def repeat(self):
        return self._repeat

    @property
    def delay_between_repeats(self):
        return self._delay_between_repeats

    @property
    def payload(self):
        """ Get the payload of the command.

        :return: The payload of the command, or None if there is no payload
        :rtype: int
        """
        return self._payload

    @payload.setter
    def payload(self, payload):
        self._payload = payload

    @property
    def is_payload(self):
        """ Determine if this command has a payload.  By default, this returns\
            True if the payload passed in to the constructor is not None, but\
            this can be overridden to indicate that a payload will be\
            generated, despite None being passed to the constructor

        :return: True if there is a payload, False otherwise
        :rtype: bool
        """
        return self._payload is not None

    def __repr__(self):
        return \
            "MultiCastCommand(time={}, key={}, payload={},"\
            " time_between_repeat={}, repeats={})".format(
                self._time, self._key, self._payload,
                self._delay_between_repeats, self._repeat)



# spinn front end common imports
from pacman.model.resources import CoreResource
from pacman.model.resources import PreAllocatedResourceContainer
from pacman.model.resources import SpecificChipSDRAMResource
from spinn_front_end_common.utility_models.live_packet_gather_machine_vertex \
    import LivePacketGatherMachineVertex as LPGVertex
from spinn_utilities.progress_bar import ProgressBar


class PreAllocateResourcesForLivePacketGatherers(object):
    """ Adds Live Packet Gatherer resources as required for a machine
    """

    def __call__(
            self, live_packet_gatherer_parameters, machine,
            pre_allocated_resources=None):
        """

        :param live_packet_gatherer_parameters:\
            the LPG parameters requested by the script
        :param previous_allocated_resources: other pre-allocated resources
        :param machine: the spinnaker machine as discovered
        :return: pre allocated resources
        """

        progress = ProgressBar(
            len(machine.ethernet_connected_chips),
            "Preallocating resources for Live Recording")

        # store how much SDRAM the LPG uses per core
        lpg_sdram_requirement = LPGVertex.get_sdram_usage()

        # for every Ethernet connected chip, get the resources needed by the
        # live packet gatherers
        sdrams = list()
        cores = list()
        for chip in progress.over(machine.ethernet_connected_chips):
            self._add_chip_lpg_reqs(live_packet_gatherer_parameters, chip,
                                    lpg_sdram_requirement, sdrams, cores)

        # create pre allocated resource container
        lpg_prealloc_resource_container = PreAllocatedResourceContainer(
            specific_sdram_usage=sdrams, core_resources=cores)

        # add other pre allocated resources
        if pre_allocated_resources is not None:
            lpg_prealloc_resource_container.extend(pre_allocated_resources)

        # return pre allocated resources
        return lpg_prealloc_resource_container

    @staticmethod
    def _add_chip_lpg_reqs(lpg_parameters, chip, lpg_sdram, sdrams, cores):
        sdram_reqs = 0
        core_reqs = 0

        for lpg_params in lpg_parameters:
            if (lpg_params.board_address is None or
                    lpg_params.board_address == chip.ip_address):
                sdram_reqs += lpg_sdram
                core_reqs += 1

        if sdram_reqs > 0:
            sdrams.append(SpecificChipSDRAMResource(chip, sdram_reqs))
        if core_reqs > 0:
            cores.append(CoreResource(chip, core_reqs))



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractReceiveBuffersToHost(object):
    """ Indicates that this object can receive buffers
    """

    __slots__ = ()

    @abstractmethod
    def get_minimum_buffer_sdram_usage(self):
        """ Get the minimum amount of SDRAM to reserve for buffers
        """

    @abstractmethod
    def get_n_timesteps_in_buffer_space(self, buffer_space, machine_time_step):
        """ Get the number of timesteps that can be stored fully in the given\
            buffer space in bytes

        :param buffer_space: The buffer space in bytes
        :param machine_time_step: The size of each time step
        :return: The number of time steps that can be stored in the buffer
        :rtype: int
        """

    @abstractmethod
    def get_recorded_region_ids(self):
        """ Get the recording region ids that have been recorded using buffering

        :return: The region numbers that have active recording
        :rtype: iterable of int
        """

    @abstractmethod
    def get_recording_region_base_address(self, txrx, placement):
        """ Get the recording region base address

        :param txrx: the SpiNNMan instance
        :param placement:\
            the placement object of the core to find the address of
        :return: the base address of the recording region
        """



from .abstract_changable_after_run import AbstractChangableAfterRun
from .abstract_generates_data_specification \
    import AbstractGeneratesDataSpecification
from .abstract_has_associated_binary import AbstractHasAssociatedBinary
from .abstract_machine_allocation_controller \
    import AbstractMachineAllocationController
from .abstract_provides_incoming_partition_constraints \
    import AbstractProvidesIncomingPartitionConstraints
from .abstract_provides_key_to_atom_mapping \
    import AbstractProvidesKeyToAtomMapping
from .abstract_provides_n_keys_for_partition \
    import AbstractProvidesNKeysForPartition
from .abstract_provides_outgoing_partition_constraints \
    import AbstractProvidesOutgoingPartitionConstraints
from .abstract_recordable import AbstractRecordable
from .abstract_rewrites_data_specification \
    import AbstractRewritesDataSpecification
from .abstract_send_me_multicast_commands_vertex \
    import AbstractSendMeMulticastCommandsVertex
from .abstract_vertex_with_dependent_vertices \
    import AbstractVertexWithEdgeToDependentVertices
from .abstract_supports_database_injection \
    import AbstractSupportsDatabaseInjection

__all__ = ["AbstractChangableAfterRun", "AbstractGeneratesDataSpecification",
           "AbstractHasAssociatedBinary",
           "AbstractMachineAllocationController",
           "AbstractProvidesIncomingPartitionConstraints",
           "AbstractProvidesKeyToAtomMapping",
           "AbstractProvidesNKeysForPartition",
           "AbstractProvidesOutgoingPartitionConstraints",
           "AbstractRecordable", "AbstractRewritesDataSpecification",
           "AbstractSendMeMulticastCommandsVertex",
           "AbstractSupportsDatabaseInjection",
           "AbstractVertexWithEdgeToDependentVertices"]



from pacman.executor.injection_decorator import inject_items
from pacman.executor.injection_decorator import supports_injection
from pacman.executor.injection_decorator import inject
from pacman.model.decorators import overrides
from pacman.model.constraints.key_allocator_constraints \
    import FixedKeyAndMaskConstraint
from pacman.model.constraints.placer_constraints import BoardConstraint
from pacman.model.resources import IPtagResource, ReverseIPtagResource
from pacman.model.resources import ResourceContainer, DTCMResource
from pacman.model.resources import SDRAMResource, CPUCyclesPerTickResource
from pacman.model.routing_info import BaseKeyAndMask
from pacman.model.graphs.machine import MachineVertex

from spinn_front_end_common.utilities.helpful_functions \
    import locate_memory_region_for_placement
from spinn_front_end_common.interface.buffer_management.recording_utilities \
    import TRAFFIC_IDENTIFIER
from spinn_front_end_common.interface.buffer_management.buffer_models \
    import SendsBuffersFromHostPreBufferedImpl, AbstractReceiveBuffersToHost
from spinn_front_end_common.interface.buffer_management.storage_objects \
    import BufferedSendingRegion
from spinn_front_end_common.utilities import constants
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.abstract_models \
    import AbstractProvidesOutgoingPartitionConstraints, AbstractRecordable
from spinn_front_end_common.abstract_models \
    import AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary
from spinn_front_end_common.abstract_models \
    import AbstractSupportsDatabaseInjection
from spinn_front_end_common.interface.simulation.simulation_utilities \
    import get_simulation_header_array
from spinn_front_end_common.interface.provenance \
    import ProvidesProvenanceDataFromMachineImpl
from spinn_front_end_common.interface.buffer_management\
    import recording_utilities
from spinn_front_end_common.utilities.utility_objs \
    import ProvenanceDataItem, ExecutableStartType

from spinnman.messages.eieio import EIEIOPrefix

from enum import Enum
import math
import sys
import struct

_DEFAULT_MALLOC_REGIONS = 2


@supports_injection
class ReverseIPTagMulticastSourceMachineVertex(
        MachineVertex, AbstractGeneratesDataSpecification,
        AbstractHasAssociatedBinary, AbstractSupportsDatabaseInjection,
        ProvidesProvenanceDataFromMachineImpl,
        AbstractProvidesOutgoingPartitionConstraints,
        SendsBuffersFromHostPreBufferedImpl,
        AbstractReceiveBuffersToHost, AbstractRecordable):
    """ A model which allows events to be injected into spinnaker and\
        converted in to multicast packets
    """

    _REGIONS = Enum(
        value="_REGIONS",
        names=[('SYSTEM', 0),
               ('CONFIGURATION', 1),
               ('RECORDING', 2),
               ('SEND_BUFFER', 3),
               ('PROVENANCE_REGION', 4)])

    _PROVENANCE_ITEMS = Enum(
        value="_PROVENANCE_ITEMS",
        names=[("N_RECEIVED_PACKETS", 0),
               ("N_SENT_PACKETS", 1),
               ("INCORRECT_KEYS", 2),
               ("INCORRECT_PACKETS", 3),
               ("LATE_PACKETS", 4)])

    # 12 ints (1. has prefix, 2. prefix, 3. prefix type, 4. check key flag,
    #          5. has key, 6. key, 7. mask, 8. buffer space,
    #          9. send buffer flag before notify, 10. tag,
    #          11. tag destination (y, x), 12. receive SDP port)
    _CONFIGURATION_REGION_SIZE = 12 * 4

    def __init__(
            self, n_keys, label, constraints=None,

            # General input and output parameters
            board_address=None,

            # Live input parameters
            receive_port=None,
            receive_sdp_port=(
                constants.SDP_PORTS.INPUT_BUFFERING_SDP_PORT.value),
            receive_tag=None,
            receive_rate=10,

            # Key parameters
            virtual_key=None, prefix=None,
            prefix_type=None, check_keys=False,

            # Send buffer parameters
            send_buffer_times=None,
            send_buffer_partition_id=None,
            send_buffer_max_space=(
                constants.MAX_SIZE_OF_BUFFERED_REGION_ON_CHIP),
            send_buffer_space_before_notify=640,

            # Buffer notification details
            buffer_notification_ip_address=None,
            buffer_notification_port=None,
            buffer_notification_tag=None,

            # Extra flag for receiving packets without a port
            reserve_reverse_ip_tag=False):
        """

        :param n_keys: The number of keys to be sent via this multicast source
        :param label: The label of this vertex
        :param constraints: Any initial constraints to this vertex
        :param board_address: The IP address of the board on which to place\
                this vertex if receiving data, either buffered or live (by\
                default, any board is chosen)
        :param receive_port: The port on the board that will listen for\
                incoming event packets (default is to disable this feature;\
                set a value to enable it)
        :param receive_sdp_port: The SDP port to listen on for incoming event\
                packets (defaults to 1)
        :param receive_tag: The IP tag to use for receiving live events\
                (uses any by default)
        :param virtual_key: The base multicast key to send received events\
                with (assigned automatically by default)
        :param prefix: The prefix to "or" with generated multicast keys\
                (default is no prefix)
        :param prefix_type: Whether the prefix should apply to the upper or\
                lower half of the multicast keys (default is upper half)
        :param check_keys: True if the keys of received events should be\
                verified before sending (default False)
        :param send_buffer_times: An array of arrays of times at which keys\
                should be sent (one array for each key, default disabled)
        :param send_buffer_max_space: The maximum amount of space to use of\
                the SDRAM on the machine (default is 1MB)
        :param send_buffer_space_before_notify: The amount of space free in\
                the sending buffer before the machine will ask the host for\
                more data (default setting is optimised for most cases)
        :param buffer_notification_ip_address: The IP address of the host\
                that will send new buffers (must be specified if a send buffer\
                is specified)
        :param buffer_notification_port: The port that the host that will\
                send new buffers is listening on (must be specified if a\
                send buffer is specified)
        :param buffer_notification_tag: The IP tag to use to notify the\
                host about space in the buffer (default is to use any tag)
        """
        MachineVertex.__init__(self, label, constraints)
        AbstractReceiveBuffersToHost.__init__(self)

        AbstractProvidesOutgoingPartitionConstraints.__init__(self)

        self._iptags = None
        self._reverse_iptags = None

        # Set up for receiving live packets
        if receive_port is not None or reserve_reverse_ip_tag:
            self._reverse_iptags = [ReverseIPtagResource(
                port=receive_port, sdp_port=receive_sdp_port,
                tag=receive_tag)]
            if board_address is not None:
                self.add_constraint(BoardConstraint(board_address))
        self._receive_rate = receive_rate
        self._receive_sdp_port = receive_sdp_port

        # Work out if buffers are being sent
        self._send_buffer = None
        self._send_buffer_partition_id = send_buffer_partition_id
        if send_buffer_times is None:
            self._send_buffer_times = None
            self._send_buffer_max_space = send_buffer_max_space
            self._send_buffers = None
        else:
            self._send_buffer_max_space = send_buffer_max_space
            self._send_buffer = BufferedSendingRegion(send_buffer_max_space)
            self._send_buffer_times = send_buffer_times

            self._iptags = [IPtagResource(
                ip_address=buffer_notification_ip_address,
                port=buffer_notification_port, strip_sdp=True,
                tag=buffer_notification_tag,
                traffic_identifier=TRAFFIC_IDENTIFIER)]
            if board_address is not None:
                self.add_constraint(BoardConstraint(board_address))
            self._send_buffers = {
                self._REGIONS.SEND_BUFFER.value:
                self._send_buffer
            }

        # buffered out parameters
        self._send_buffer_space_before_notify = send_buffer_space_before_notify
        if self._send_buffer_space_before_notify > send_buffer_max_space:
            self._send_buffer_space_before_notify = send_buffer_max_space

        # Set up for recording (if requested)
        self._record_buffer_size = 0
        self._buffer_size_before_receive = 0
        self._time_between_triggers = 0
        self._maximum_recording_buffer = 0

        # Set up for buffering
        self._buffer_notification_ip_address = buffer_notification_ip_address
        self._buffer_notification_port = buffer_notification_port
        self._buffer_notification_tag = buffer_notification_tag

        # set flag for checking if in injection mode
        self._in_injection_mode = receive_port is not None

        # Sort out the keys to be used
        self._n_keys = n_keys
        self._virtual_key = virtual_key
        self._mask = None
        self._prefix = prefix
        self._prefix_type = prefix_type
        self._check_keys = check_keys

        # Work out the prefix details
        if self._prefix is not None:
            if self._prefix_type is None:
                self._prefix_type = EIEIOPrefix.UPPER_HALF_WORD
            if self._prefix_type == EIEIOPrefix.UPPER_HALF_WORD:
                self._prefix = prefix << 16

        # If the user has specified a virtual key
        if self._virtual_key is not None:

            # check that virtual key is valid
            if self._virtual_key < 0:
                raise ConfigurationException(
                    "Virtual keys must be positive")

            # Get a mask and maximum number of keys for the number of keys
            # requested
            self._mask, max_key = self._calculate_mask(n_keys)

            # Check that the number of keys and the virtual key don't interfere
            if n_keys > max_key:
                raise ConfigurationException(
                    "The mask calculated from the number of keys will "
                    "not work with the virtual key specified")

            if self._prefix is not None:

                # Check that the prefix doesn't change the virtual key in the
                # masked area
                masked_key = (self._virtual_key | self._prefix) & self._mask
                if self._virtual_key != masked_key:
                    raise ConfigurationException(
                        "The number of keys, virtual key and key prefix"
                        " settings don't work together")
            else:

                # If no prefix was generated, generate one
                self._prefix_type = EIEIOPrefix.UPPER_HALF_WORD
                self._prefix = self._virtual_key

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self):
        return self._REGIONS.PROVENANCE_REGION.value

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self):
        return 5

    @property
    @overrides(MachineVertex.resources_required)
    def resources_required(self):
        resources = ResourceContainer(
            dtcm=DTCMResource(self.get_dtcm_usage()),
            sdram=SDRAMResource(self.get_sdram_usage(
                self._send_buffer_times, self._send_buffer_max_space,
                self._record_buffer_size > 0)),
            cpu_cycles=CPUCyclesPerTickResource(self.get_cpu_usage()),
            iptags=self._iptags,
            reverse_iptags=self._reverse_iptags)
        if self._iptags is None:
            resources.extend(recording_utilities.get_recording_resources(
                [self._record_buffer_size],
                self._buffer_notification_ip_address,
                self._buffer_notification_port, self._buffer_notification_tag))
        else:
            resources.extend(recording_utilities.get_recording_resources(
                [self._record_buffer_size]))
        return resources

    @staticmethod
    def get_sdram_usage(
            send_buffer_times, send_buffer_max_space, recording_enabled):
        send_buffer_size = 0
        if send_buffer_times is not None:
            send_buffer_size = send_buffer_max_space

        mallocs = \
            ReverseIPTagMulticastSourceMachineVertex.n_regions_to_allocate(
                send_buffer_times is not None, recording_enabled)
        allocation_size = mallocs * constants.SARK_PER_MALLOC_SDRAM_USAGE

        return (
            constants.SYSTEM_BYTES_REQUIREMENT +
            (ReverseIPTagMulticastSourceMachineVertex.
                _CONFIGURATION_REGION_SIZE) +
            send_buffer_size + allocation_size +
            (ReverseIPTagMulticastSourceMachineVertex.
                get_provenance_data_size(0)))

    @staticmethod
    def get_dtcm_usage():
        return 1

    @staticmethod
    def get_cpu_usage():
        return 1

    @staticmethod
    def n_regions_to_allocate(send_buffering, recording):
        """ Get the number of regions that will be allocated
        """
        if recording and send_buffering:
            return 5
        elif recording or send_buffering:
            return 4
        return 3

    @property
    def send_buffer_times(self):
        return self._send_buffer_times

    @send_buffer_times.setter
    def send_buffer_times(self, send_buffer_times):
        self._send_buffer_times = send_buffer_times

    def _is_in_range(
            self, time_stamp_in_ticks,
            first_machine_time_step, n_machine_time_steps):
        return (
            (n_machine_time_steps is None) or (
                first_machine_time_step <= time_stamp_in_ticks <
                n_machine_time_steps))

    def _fill_send_buffer(
            self, machine_time_step, first_machine_time_step,
            n_machine_time_steps):
        """ Fill the send buffer with keys to send
        """

        key_to_send = self._virtual_key
        if self._virtual_key is None:
            key_to_send = 0

        if self._send_buffer is not None:
            self._send_buffer.clear()
        if (self._send_buffer_times is not None and
                len(self._send_buffer_times) != 0):
            if hasattr(self._send_buffer_times[0], "__len__"):

                # Works with a list-of-lists
                for key in range(self._n_keys):
                    for timeStamp in sorted(self._send_buffer_times[key]):
                        time_stamp_in_ticks = int(math.ceil(
                            float(int(timeStamp * 1000.0)) /
                            machine_time_step))
                        if self._is_in_range(
                                time_stamp_in_ticks, first_machine_time_step,
                                n_machine_time_steps):
                            self._send_buffer.add_key(
                                time_stamp_in_ticks, key_to_send + key)
            else:

                # Work with a single list
                key_list = [
                    key + key_to_send for key in range(self._n_keys)]
                for timeStamp in sorted(self._send_buffer_times):
                    time_stamp_in_ticks = int(math.ceil(
                        float(int(timeStamp * 1000.0)) /
                        machine_time_step))

                    # add to send_buffer collection
                    if self._is_in_range(
                            time_stamp_in_ticks, first_machine_time_step,
                            n_machine_time_steps):
                        self._send_buffer.add_keys(
                            time_stamp_in_ticks, key_list)

    @staticmethod
    def _generate_prefix(virtual_key, prefix_type):
        if prefix_type == EIEIOPrefix.LOWER_HALF_WORD:
            return virtual_key & 0xFFFF
        return (virtual_key >> 16) & 0xFFFF

    @staticmethod
    def _calculate_mask(n_neurons):
        if n_neurons == 1:
            return 0xFFFFFFFF, 1
        temp_value = int(math.ceil(math.log(n_neurons, 2)))
        max_key = (int(math.pow(2, temp_value)) - 1)
        mask = 0xFFFFFFFF - max_key
        return mask, max_key

    def enable_recording(
            self,
            record_buffer_size=constants.MAX_SIZE_OF_BUFFERED_REGION_ON_CHIP,
            buffer_size_before_receive=(
                constants.DEFAULT_BUFFER_SIZE_BEFORE_RECEIVE),
            time_between_triggers=0):
        """ Enable recording of the keys sent

        :param record_buffer_size:\
            The size of the recording buffer in bytes.  Note that when using\
            automatic pause and resume, this will be used as the minimum size\
            of the buffer
        :type record_buffer_size: int
        :param buffer_size_before_receive:\
            The size that the buffer can grow to before a read request is\
            issued to the host (in bytes)
        :type buffer_size_before_receive: int
        :param time_between_triggers:\
            The minimum time between the sending of read requests
        :type time_between_triggers: int
        """
        self._record_buffer_size = record_buffer_size
        self._buffer_size_before_receive = buffer_size_before_receive
        self._time_between_triggers = time_between_triggers

    def _reserve_regions(self, spec):

        # Reserve system and configuration memory regions:
        spec.reserve_memory_region(
            region=self._REGIONS.SYSTEM.value,
            size=constants.SYSTEM_BYTES_REQUIREMENT, label='SYSTEM')
        spec.reserve_memory_region(
            region=self._REGIONS.CONFIGURATION.value,
            size=self._CONFIGURATION_REGION_SIZE, label='CONFIGURATION')

        # Reserve recording buffer regions if required
        spec.reserve_memory_region(
            region=self._REGIONS.RECORDING.value,
            size=recording_utilities.get_recording_header_size(1),
            label="RECORDING")

        # Reserve send buffer region if required
        if self._send_buffer_times is not None:
            max_buffer_size = self.get_max_buffer_size_possible(
                self._REGIONS.SEND_BUFFER.value)
            spec.reserve_memory_region(
                region=self._REGIONS.SEND_BUFFER.value,
                size=max_buffer_size, label="SEND_BUFFER", empty=True)

        self.reserve_provenance_data_region(spec)

    def _update_virtual_key(self, routing_info, machine_graph):
        if self._virtual_key is None:
            if self._send_buffer_partition_id is not None:

                rinfo = routing_info.get_routing_info_from_pre_vertex(
                    self, self._send_buffer_partition_id)

                # if no edge leaving this vertex, no key needed
                if rinfo is not None:
                    self._virtual_key = rinfo.first_key
                    self._mask = rinfo.first_mask

            else:
                partitions = machine_graph\
                    .get_outgoing_edge_partitions_starting_at_vertex(self)
                partition = next(iter(partitions), None)

                if partition is not None:
                    rinfo = routing_info.get_routing_info_from_partition(
                        partition)
                    self._virtual_key = rinfo.first_key
                    self._mask = rinfo.first_mask

        if self._virtual_key is not None and self._prefix is None:
            self._prefix_type = EIEIOPrefix.UPPER_HALF_WORD
            self._prefix = self._virtual_key

    def _write_configuration(self, spec, ip_tags):
        spec.switch_write_focus(region=self._REGIONS.CONFIGURATION.value)

        # Write apply_prefix and prefix and prefix_type
        if self._prefix is None:
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)
        else:
            spec.write_value(data=1)
            spec.write_value(data=self._prefix)
            spec.write_value(data=self._prefix_type.value)

        # Write check
        if self._check_keys:
            spec.write_value(data=1)
        else:
            spec.write_value(data=0)

        # Write if you have a key to transmit write it and the mask,
        # otherwise write flags to fill in space
        if self._virtual_key is None:
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)
        else:
            spec.write_value(data=1)
            spec.write_value(data=self._virtual_key)
            spec.write_value(data=self._mask)

        # Write send buffer data
        if self._send_buffer_times is not None:

            this_tag = None

            for tag in ip_tags:
                if tag.traffic_identifier == TRAFFIC_IDENTIFIER:
                    this_tag = tag
                    break
            if this_tag is None:
                raise Exception("Could not find tag for send buffering")

            buffer_space = self.get_max_buffer_size_possible(
                self._REGIONS.SEND_BUFFER.value)

            spec.write_value(data=buffer_space)
            spec.write_value(data=self._send_buffer_space_before_notify)
            spec.write_value(data=this_tag.tag)
            spec.write_value(struct.unpack("<I", struct.pack(
                "<HH", this_tag.destination_y, this_tag.destination_x))[0])
        else:
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)
            spec.write_value(data=0)

        # write SDP port to which SDP packets will be received
        spec.write_value(data=self._receive_sdp_port)

    @inject_items({
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor",
        "machine_graph": "MemoryMachineGraph",
        "routing_info": "MemoryRoutingInfos",
        "tags": "MemoryTags",
        "first_machine_time_step": "FirstMachineTimeStep",
        "n_machine_time_steps": "TotalMachineTimeSteps"
    })
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_time_step", "time_scale_factor", "machine_graph",
            "routing_info", "tags", "first_machine_time_step",
            "n_machine_time_steps"
        })
    def generate_data_specification(
            self, spec, placement, machine_time_step, time_scale_factor,
            machine_graph, routing_info, tags, first_machine_time_step,
            n_machine_time_steps):

        self._update_virtual_key(routing_info, machine_graph)
        self._fill_send_buffer(
            machine_time_step, first_machine_time_step, n_machine_time_steps)

        # Reserve regions
        self._reserve_regions(spec)

        # Write the system region
        spec.switch_write_focus(self._REGIONS.SYSTEM.value)
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name(), machine_time_step,
            time_scale_factor))

        # Write the additional recording information
        iptags = tags.get_ip_tags_for_vertex(self)
        spec.switch_write_focus(self._REGIONS.RECORDING.value)
        spec.write_array(recording_utilities.get_recording_header_array(
            [self._record_buffer_size],
            self._time_between_triggers, self._buffer_size_before_receive,
            iptags, self._buffer_notification_tag))

        # Write the configuration information
        self._write_configuration(spec, iptags)

        # End spec
        spec.end_specification()

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return "reverse_iptag_multicast_source.aplx"

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableStartType.USES_SIMULATION_INTERFACE

    @overrides(AbstractProvidesOutgoingPartitionConstraints.
               get_outgoing_partition_constraints)
    def get_outgoing_partition_constraints(self, partition):
        if self._virtual_key is not None:
            return list([FixedKeyAndMaskConstraint(
                [BaseKeyAndMask(self._virtual_key, self._mask)])])
        return list()

    @property
    def virtual_key(self):
        return self._virtual_key

    @property
    def mask(self):
        return self._mask

    @property
    @overrides(AbstractSupportsDatabaseInjection.is_in_injection_mode)
    def is_in_injection_mode(self):
        return self._in_injection_mode

    @overrides(AbstractRecordable.is_recording)
    def is_recording(self):
        return self._record_buffer_size > 0

    @inject("FirstMachineTimeStep")
    @inject_items({
        "machine_time_step": "MachineTimeStep",
        "n_machine_time_steps": "TotalMachineTimeSteps"
    })
    def update_buffer(
            self, first_machine_time_step, machine_time_step,
            n_machine_time_steps):
        self._fill_send_buffer(
            machine_time_step, first_machine_time_step, n_machine_time_steps)

    @overrides(AbstractReceiveBuffersToHost.get_minimum_buffer_sdram_usage)
    def get_minimum_buffer_sdram_usage(self):
        return self._record_buffer_size

    @overrides(AbstractReceiveBuffersToHost.get_n_timesteps_in_buffer_space)
    def get_n_timesteps_in_buffer_space(self, buffer_space, machine_time_step):

        # If not recording, not an issue
        if self._record_buffer_size == 0:
            return sys.maxint

        # If recording and using pre-defined keys, use the maximum
        if self._send_buffer is not None:
            return recording_utilities.get_n_timesteps_in_buffer_space(
                buffer_space, [self._send_buffer.max_packets_in_timestamp])

        # If recording and not using pre-defined keys, use the specified
        # rate to work it out - add 10% for safety
        keys_per_timestep = math.ceil(
            (self._receive_rate / (machine_time_step * 1000.0)) * 1.1
        )

        # 4 bytes per key + 2 byte header + 4 byte timestamp
        bytes_per_timestep = (keys_per_timestep * 4) + 6
        return recording_utilities.get_n_timesteps_in_buffer_space(
            buffer_space, [bytes_per_timestep])

    @overrides(AbstractReceiveBuffersToHost.get_recorded_region_ids)
    def get_recorded_region_ids(self):
        return recording_utilities.get_recorded_region_ids(
            [self._record_buffer_size])

    @overrides(AbstractReceiveBuffersToHost.get_recording_region_base_address)
    def get_recording_region_base_address(self, txrx, placement):
        return locate_memory_region_for_placement(
            placement, self._REGIONS.RECORDING.value, txrx)

    @property
    def send_buffers(self):
        return self._send_buffers

    @send_buffers.setter
    def send_buffers(self, value):
        self._send_buffers = value

    @overrides(ProvidesProvenanceDataFromMachineImpl.
               get_provenance_data_from_machine)
    def get_provenance_data_from_machine(self, transceiver, placement):
        provenance_data = self._read_provenance_data(transceiver, placement)
        provenance_items = self._read_basic_provenance_items(
            provenance_data, placement)
        provenance_data = self._get_remaining_provenance_data_items(
            provenance_data)
        _, _, _, _, names = self._get_placement_details(placement)

        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "received_sdp_packets"),
            provenance_data[self._PROVENANCE_ITEMS.N_RECEIVED_PACKETS.value],
            report=(
                provenance_data[
                    self._PROVENANCE_ITEMS.N_RECEIVED_PACKETS.value] == 0 and
                self._send_buffer_times is None),
            message=(
                "No SDP packets were received by {}.  If you expected packets"
                " to be injected, this could indicate an error".format(
                    self._label))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "send_multicast_packets"),
            provenance_data[self._PROVENANCE_ITEMS.N_SENT_PACKETS.value],
            report=provenance_data[
                self._PROVENANCE_ITEMS.N_SENT_PACKETS.value] == 0,
            message=(
                "No multicast packets were sent by {}.  If you expected"
                " packets to be sent this could indicate an error".format(
                    self._label))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "incorrect_keys"),
            provenance_data[self._PROVENANCE_ITEMS.INCORRECT_KEYS.value],
            report=provenance_data[
                self._PROVENANCE_ITEMS.INCORRECT_KEYS.value] > 0,
            message=(
                "Keys were received by {} that did not match the key {} and"
                " mask {}".format(
                    self._label, self._virtual_key, self._mask))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "incorrect_packets"),
            provenance_data[self._PROVENANCE_ITEMS.INCORRECT_PACKETS.value],
            report=provenance_data[
                self._PROVENANCE_ITEMS.INCORRECT_PACKETS.value] > 0,
            message=(
                "SDP Packets were received by {} that were not correct".format(
                    self._label))))
        provenance_items.append(ProvenanceDataItem(
            self._add_name(names, "late_packets"),
            provenance_data[self._PROVENANCE_ITEMS.LATE_PACKETS.value],
            report=provenance_data[
                self._PROVENANCE_ITEMS.LATE_PACKETS.value] > 0,
            message=(
                "SDP Packets were received by {} that were too late to be"
                " transmitted in the simulation".format(self._label))))

        return provenance_items

    def __repr__(self):
        return self._label



# spinnman imports
from spinnman.connections import SocketAddressWithChip
from spinnman.transceiver import create_transceiver_from_hostname
from spinnman.model import BMPConnectionData

# front end common imports
from spinn_front_end_common.utilities.exceptions import ConfigurationException

# general imports
import re


class MachineGenerator(object):
    """ Interface to make a transceiver and a spinn_machine object
    """

    __slots__ = []

    def __call__(
            self, hostname, bmp_details, downed_chips, downed_cores,
            downed_links, board_version, auto_detect_bmp, enable_reinjection,
            scamp_connection_data, boot_port_num, reset_machine_on_start_up,
            max_sdram_size=None, max_core_id=None):
        """
        :param hostname: the hostname or ip address of the spinnaker machine
        :param bmp_details: the details of the BMP connections
        :param downed_chips: the chips that are down which SARK thinks are\
                alive
        :param downed_cores: the cores that are down which SARK thinks are\
                alive
        :param board_version: the version of the boards being used within the\
                machine (1, 2, 3, 4 or 5)
        :param auto_detect_bmp: boolean which determines if the BMP should
               be automatically determined
        :param enable_reinjection: True if dropped packet reinjection is to be\
               enabled
        :param boot_port_num: the port num used for the boot connection
        :param scamp_connection_data: the list of scamp connection data or\
               None
        :param max_sdram_size: the maximum SDRAM each chip can say it has
               (mainly used in debugging purposes)
        :type max_sdram_size: int or None
        """

        # if the end user gives you scamp data, use it and don't discover them
        if scamp_connection_data is not None:
            scamp_connection_data = \
                self._sort_out_scamp_connections(scamp_connection_data)

        # sort out BMP connections into list of strings
        bmp_connection_data = self._sort_out_bmp_string(bmp_details)

        txrx = create_transceiver_from_hostname(
            hostname=hostname, bmp_connection_data=bmp_connection_data,
            version=board_version, ignore_chips=downed_chips,
            ignore_cores=downed_cores, ignored_links=downed_links,
            auto_detect_bmp=auto_detect_bmp,
            boot_port_no=boot_port_num,
            scamp_connections=scamp_connection_data,
            max_sdram_size=max_sdram_size, max_core_id=max_core_id)

        if reset_machine_on_start_up:
            txrx.power_off_machine()

        # do auto boot if possible
        if board_version is None:
            raise ConfigurationException(
                "Please set a machine version number in the configuration "
                "file (spynnaker.cfg or pacman.cfg)")
        txrx.ensure_board_is_ready(
            enable_reinjector=enable_reinjection)
        txrx.discover_scamp_connections()
        machine = txrx.get_machine_details()
        return machine, txrx

    @staticmethod
    def _sort_out_scamp_connections(scamp_connections_data):
        scamp_addresses = list()
        for scamp_connection in scamp_connections_data.split(":"):
            scamp_connection_split = scamp_connection.split(",")
            if len(scamp_connection_split) == 3:
                scamp_addresses.append(SocketAddressWithChip(
                    hostname=scamp_connection_split[0],
                    port_num=None,
                    chip_x=int(scamp_connection_split[1]),
                    chip_y=int(scamp_connection_split[2])))
            else:
                scamp_addresses.append(SocketAddressWithChip(
                    hostname=scamp_connection_split[0],
                    port_num=int(scamp_connection_split[1]),
                    chip_x=int(scamp_connection_split[2]),
                    chip_y=int(scamp_connection_split[3])))
        return scamp_addresses

    @staticmethod
    def _sort_out_bmp_cabinet_and_frame_string(bmp_cabinet_and_frame):
        split_string = bmp_cabinet_and_frame.split(";", 2)
        if len(split_string) == 1:
            hostname_split = split_string[0].split(",")
            if len(hostname_split) == 1:
                return [0, 0, split_string[0], None]
            else:
                return [0, 0, hostname_split[0], hostname_split[1]]
        if len(split_string) == 2:
            hostname_split = split_string[1].split(",")
            if len(hostname_split) == 1:
                return [0, split_string[0], split_string[1], None]
            else:
                return [0, split_string[0], hostname_split[0],
                        hostname_split[1]]
        hostname_split = split_string[2].split(",")
        if len(hostname_split) == 1:
            return [split_string[0], split_string[1], hostname_split[0], None]
        else:
            return [split_string[0], split_string[1], hostname_split[0],
                    hostname_split[1]]

    @staticmethod
    def _sort_out_bmp_boards_string(bmp_boards):

        # If the string is a range of boards, get the range
        range_match = re.match("(\d+)-(\d+)", bmp_boards)
        if range_match is not None:
            return range(int(range_match.group(1)),
                         int(range_match.group(2)) + 1)

        # Otherwise, assume a list of boards
        return [int(board) for board in bmp_boards.split(",")]

    def _sort_out_bmp_string(self, bmp_string):
        """ Take a BMP line and split it into the BMP connection data

        :param bmp_string: the BMP string to be converted
        :return: the BMP connection data
        """
        bmp_details = list()
        if bmp_string is None or bmp_string == "None":
            return None

        for bmp_detail in bmp_string.split(":"):

            bmp_string_split = bmp_detail.split("/")
            (cabinet, frame, hostname, port_num) = \
                self._sort_out_bmp_cabinet_and_frame_string(
                    bmp_string_split[0])

            if len(bmp_string_split) == 1:

                # if there is no split, then assume its one board,
                # located at position 0
                if port_num is not None:
                    bmp_details.append(
                        BMPConnectionData(cabinet, frame, hostname, [0],
                                          int(port_num)))
                else:
                    bmp_details.append(
                        BMPConnectionData(cabinet, frame, hostname, [0],
                                          port_num=None))
            else:
                boards = self._sort_out_bmp_boards_string(bmp_string_split[1])
                if port_num is not None:
                    bmp_details.append(
                        BMPConnectionData(cabinet, frame, hostname, boards,
                                          int(port_num)))
                else:
                    bmp_details.append(
                        BMPConnectionData(cabinet, frame, hostname, boards,
                                          None))
        return bmp_details



from pacman.model.decorators import overrides
from pacman.executor.injection_decorator \
    import supports_injection, inject_items

from spinn_front_end_common.abstract_models \
    import AbstractGeneratesDataSpecification

from abc import abstractmethod


@supports_injection
class MachineDataSpecableVertex(AbstractGeneratesDataSpecification):

    __slots__ = ()

    @inject_items({
        "machine_graph": "MemoryMachineGraph",
        "routing_info": "MemoryRoutingInfos",
        "tags": "MemoryTags",
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor"
    })
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_graph", "routing_info", "tags",
            "machine_time_step", "time_scale_factor"
        })
    def generate_data_specification(
            self, spec, placement, machine_graph, routing_info, tags,
            machine_time_step, time_scale_factor):
        iptags = tags.get_ip_tags_for_vertex(placement.vertex)
        reverse_iptags = tags.get_reverse_ip_tags_for_vertex(placement.vertex)
        self.generate_machine_data_specification(
            spec, placement, machine_graph, routing_info, iptags,
            reverse_iptags, machine_time_step, time_scale_factor)

    @abstractmethod
    def generate_machine_data_specification(
            self, spec, placement, machine_graph, routing_info, iptags,
            reverse_iptags, machine_time_step, time_scale_factor):
        pass



from data_specification import data_spec_sender

from spinn_utilities.progress_bar import ProgressBar
from spinn_machine import CoreSubsets

# spinnman imports
from spinnman.model.enums import CPUState

# front end common imports
from spinn_front_end_common.utilities.constants import DSE_DATA_STRUCT_SIZE

import os
import logging
import struct

logger = logging.getLogger(__name__)


class MachineExecuteDataSpecification(object):
    """ Executes the machine based data specification
    """

    __slots__ = []

    def __call__(
            self, write_memory_map_report, dsg_targets, transceiver, app_id):
        """
        :param write_memory_map_report:
        :param dsg_targets:
        :param transceiver:
        :param app_id:
        """
        return self.spinnaker_based_data_specification_execution(
            write_memory_map_report, dsg_targets, transceiver, app_id)

    def spinnaker_based_data_specification_execution(
            self, write_memory_map_report, dsg_targets, transceiver, app_id):
        """

        :param write_memory_map_report:
        :param dsg_targets:
        :param transceiver:
        :param app_id:
        :return: True
        :rtype: bool
        """

        # create a progress bar for end users
        progress = ProgressBar(dsg_targets, "Loading data specifications")

        dse_app_id = transceiver.app_id_tracker.get_new_id()

        core_subset = CoreSubsets()
        for (x, y, p, label) in progress.over(dsg_targets):
            core_subset.add_processor(x, y, p)

            dse_data_struct_address = transceiver.malloc_sdram(
                x, y, DSE_DATA_STRUCT_SIZE, dse_app_id)

            data_spec_file_path = dsg_targets[x, y, p, label]
            data_spec_file_size = os.path.getsize(data_spec_file_path)

            base_address = transceiver.malloc_sdram(
                x, y, data_spec_file_size, dse_app_id)

            dse_data_struct_data = struct.pack(
                "<4I", base_address, data_spec_file_size, app_id,
                write_memory_map_report)

            transceiver.write_memory(
                x, y, dse_data_struct_address, dse_data_struct_data,
                len(dse_data_struct_data))

            transceiver.write_memory(
                x, y, base_address, data_spec_file_path, is_filename=True)

            # data spec file is written at specific address (base_address)
            # this is encapsulated in a structure with four fields:
            # 1 - data specification base address
            # 2 - data specification file size
            # 3 - future application ID
            # 4 - store data for memory map report (True / False)
            # If the memory map report is going to be produced, the
            # address of the structure is returned in user1
            user_0_address = transceiver.\
                get_user_0_register_address_from_core(x, y, p)

            transceiver.write_memory(
                x, y, user_0_address, dse_data_struct_address, 4)

        # Execute the DSE on all the cores
        logger.info("Loading the Data Specification Executor")
        dse_exec = os.path.join(
            os.path.dirname(data_spec_sender),
            'data_specification_executor.aplx')
        transceiver.execute_flood(
            core_subset, dse_exec, app_id, is_filename=True)

        logger.info(
            "Waiting for On-chip Data Specification Executor to complete")
        transceiver.wait_for_cores_to_be_in_state(
            core_subset, app_id, [CPUState.FINISHED])

        transceiver.stop_application(dse_app_id)
        transceiver.app_id_tracker.free_id(dse_app_id)
        logger.info("On-chip Data Specification Executor completed")

        return True



from pacman.model.graphs.application import ApplicationEdge
from pacman.model.graphs.machine import MachineEdge
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_utilities.progress_bar import ProgressBar


class InsertEdgesToLivePacketGatherers(object):
    """ Add edges from the recorded vertices to the local Live PacketGatherers
    """

    def __call__(
            self, live_packet_gatherer_parameters, placements,
            live_packet_gatherers_to_vertex_mapping, machine,
            machine_graph, application_graph=None, graph_mapper=None):
        """

        :param live_packet_gatherer_parameters: the set of parameters
        :param placements: the placements object
        :param live_packet_gatherers_to_vertex_mapping:\
            the mapping of LPG parameters and the machine vertices associated\
            with it
        :param machine: the SpiNNaker machine
        :param machine_graph: the machine graph
        :param application_graph:  the app graph
        :param graph_mapper: the graph mapper between app and machine graph
        :rtype: None
        """

        progress = ProgressBar(
            live_packet_gatherer_parameters,
            string_describing_what_being_progressed=(
                "Adding edges to the machine graph between the vertices to "
                "which live output has been requested and its local Live "
                "Packet Gatherer"))

        for lpg_params in progress.over(live_packet_gatherer_parameters):
            # locate vertices needed to be connected to a LPG with these params
            for vertex in live_packet_gatherer_parameters[lpg_params]:
                self._connect_lpg_vertex(
                    application_graph, graph_mapper, machine,
                    placements, machine_graph, vertex,
                    live_packet_gatherers_to_vertex_mapping, lpg_params)

    def _connect_lpg_vertex(
            self, app_graph, mapper, machine, placements, m_graph, vertex,
            lpg_to_vertex, lpg_params):
        # Find all Live Gatherer machine vertices
        m_lpgs = lpg_to_vertex[lpg_params]

        # If it is an application graph, find the machine vertices
        if app_graph is not None:
            # flag for ensuring we don't add the edge to the app
            # graph many times
            app_graph_edge = None

            # iterate through the associated machine vertices
            machine_vertices = mapper.get_machine_vertices(vertex)
            for machine_vertex in machine_vertices:
                # add a edge between the closest LPG and the vertex
                machine_edge, machine_lpg = self._process_m_vertex(
                    machine_vertex, m_lpgs, machine, placements,
                    m_graph, lpg_params.partition_id)

                # update the app graph and graph mapper
                app_graph_edge = self._update_app_graph_and_mapper(
                    app_graph, mapper, machine_lpg, vertex,
                    lpg_params.partition_id, machine_edge, app_graph_edge)
        else:
            # add a edge between the closest LPG and the vertex
            self._process_m_vertex(
                vertex, m_lpgs, machine, placements,
                m_graph, lpg_params.partition_id)

    def _process_m_vertex(
            self, machine_vertex, m_lpgs, machine,
            placements, machine_graph, partition_id):
        """ locates and places an edge for a machine vertex

        :param machine_vertex: the machine vertex that needs an edge to a LPG
        :param m_lpgs:\
            dict of chip placed on to gatherers that are associated with the\
            parameters
        :param machine: the spinnaker machine object
        :param placements: the placements object
        :param machine_graph: the machine graph object
        :param partition_id: the partition id to add to the edge
        :return: machine edge and the LPG vertex
        """

        # locate the LPG that's closest to this vertex
        machine_lpg = self._find_closest_live_packet_gatherer(
            machine_vertex, m_lpgs, machine, placements)

        # add edge for the machine graph
        machine_edge = MachineEdge(machine_vertex, machine_lpg)
        machine_graph.add_edge(machine_edge, partition_id)

        # return the machine edge
        return machine_edge, machine_lpg

    @staticmethod
    def _update_app_graph_and_mapper(
            application_graph, graph_mapper, machine_lpg, vertex,
            partition_id, machine_edge, app_graph_edge):
        """ handles changes to the app graph and graph mapper.

        :param application_graph: the app graph
        :param graph_mapper: the graph mapper
        :param machine_lpg: the machine LPG
        :param vertex: the app vertex to link to
        :param partition_id: the partition id to put the edge on
        :return the application edge for this vertex and LPG
        :rtype: ApplicationEdge
        """

        # locate app vertex for LPG
        lpg_app_vertex = graph_mapper.get_application_vertex(machine_lpg)

        # if not built the app edge, add the app edge now
        if app_graph_edge is None:
            app_graph_edge = ApplicationEdge(vertex, lpg_app_vertex)
            application_graph.add_edge(app_graph_edge, partition_id)

        # add mapping between the app edge and the machine edge
        graph_mapper.add_edge_mapping(machine_edge, app_graph_edge)

        # return the app edge for reuse as needed
        return app_graph_edge

    @staticmethod
    def _find_closest_live_packet_gatherer(
            machine_vertex, machine_lpgs, machine, placements):
        """ locates the LPG on the nearest Ethernet connected chip to the\
            machine vertex in question, or the LPG on 0, 0 if a closer one\
            can't be found

        :param machine_vertex: the machine vertex to locate the nearest LPG to
        :param machine_lpgs: dict of gatherers by chip placed on
        :param machine: the spinn machine object
        :param placements: the placements object
        :return: the local LPG
        :raise ConfigurationException: if a local gatherer cannot be found
        """

        # locate location of vertex in machine
        placement = placements.get_placement_of_vertex(machine_vertex)
        chip = machine.get_chip_at(placement.x, placement.y)

        # locate closest LPG
        if (chip.nearest_ethernet_x, chip.nearest_ethernet_y) in machine_lpgs:
            return machine_lpgs[
                chip.nearest_ethernet_x, chip.nearest_ethernet_y]

        if (0, 0) in machine_lpgs:
            return machine_lpgs[0, 0]

        # if got through all LPG vertices and not found the right one. go BOOM
        raise ConfigurationException(
            "Cannot find a Live Packet Gatherer from {} for the vertex {}"
            " located {}:{}".format(
                machine_lpgs, machine_vertex, chip.x, chip.y))



from spinn_front_end_common.utilities.constants import BUFFERING_OPERATIONS

import struct


class ChannelBufferState(object):
    """ Stores information related to a single channel output\
        buffering state, as it is retrieved at the end of a simulation on the\
        SpiNNaker system.
    """

    __slots__ = [
        # start buffering area memory address (32 bits)
        "_start_address",

        # address where data was last written (32 bits)
        "_current_write",

        # address where the dma write got up to (32 bits)
        "_current_dma_write",

        # address where data was last read (32 bits)
        "_current_read",

        # The address of first byte after the buffer (32 bits)
        "_end_address",

        # The id of the region (8 bits)
        "_region_id",

        # True if the region overflowed during the simulation (8 bits)
        "_missing_info",

        # Last operation performed on the buffer - read or write (8 bits)
        "_last_buffer_operation",

        # bool check for if its extracted data from machine
        "_update_completed",
    ]

    # 4 for _start_address, 4 for _current_write, 4 for current_dma_write,
    # 4 for _current_read, 4 for _end_address, 1 for _region_id,
    # 1 for _missing_info, 1 for _last_buffer_operation,
    ChannelBufferStateSize = 24

    def __init__(
            self, start_address, current_write, current_dma_write,
            current_read, end_address, region_id, missing_info,
            last_buffer_operation):
        """

        :param start_address: start buffering area memory address (32 bits)
        :param current_write: address where data was last written (32 bits)
        :param current_read: address where data was last read (32 bits)
        :param end_address: The address of first byte after the buffer\
                (32 bits)
        :param region_id: The id of the region (8 bits)
        :param missing_info: True if the region overflowed during the\
                simulation (8 bits)
        :param last_buffer_operation: Last operation performed on the buffer\
                - read or write (8 bits)
        """
        self._start_address = start_address
        self._current_write = current_write
        self._current_dma_write = current_dma_write
        self._current_read = current_read
        self._end_address = end_address
        self._region_id = region_id
        self._missing_info = missing_info
        self._last_buffer_operation = last_buffer_operation
        self._update_completed = False

    @property
    def start_address(self):
        return self._start_address

    @property
    def current_write(self):
        return self._current_write

    @property
    def current_read(self):
        return self._current_read

    @property
    def end_address(self):
        return self._end_address

    @property
    def region_id(self):
        return self._region_id

    @property
    def missing_info(self):
        return self._missing_info

    @property
    def last_buffer_operation(self):
        return self._last_buffer_operation

    @property
    def is_state_updated(self):
        return self._update_completed

    def update_last_operation(self, operation):
        self._last_buffer_operation = operation

    def update_read_pointer(self, read_ptr):
        self._current_read = read_ptr

    def set_update_completed(self):
        self._update_completed = True

    @staticmethod
    def create_from_bytearray(data):
        (start_address, current_write, current_dma_write, current_read,
         end_address, region_id, missing_info, last_buffer_operation) = \
            struct.unpack_from("<IIIIIBBBx", data)
        if last_buffer_operation == 0:
            last_buffer_operation = BUFFERING_OPERATIONS.BUFFER_READ.value
        else:
            last_buffer_operation = BUFFERING_OPERATIONS.BUFFER_WRITE.value
        buffer_state = ChannelBufferState(
            start_address, current_write, current_dma_write, current_read,
            end_address, region_id, missing_info, last_buffer_operation)
        return buffer_state

    @staticmethod
    def size_of_channel_state():
        return ChannelBufferState.ChannelBufferStateSize



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractProvidesNKeysForPartition(object):
    """ Allows a vertex to provide the number of keys for a partition of edges,\
        rather than relying on the number of atoms in the pre-vertex
    """

    __slots__ = ()

    @abstractmethod
    def get_n_keys_for_partition(self, partition, graph_mapper):
        """ Get the number of keys required by the given partition of edges

        :param partition: An partition that comes out of this vertex
        :type partition:\
                    :py:class:`pacman.utilities.utility_objs.outgoing_partition.OutgoingPartition`
        :param graph_mapper: A mapper between the graphs
        :type graph_mapper:\
                    :py:class:`pacman.model.graph.graph_mapper.GraphMapper`
        :return: A list of constraints
        :rtype: list of\
                    :py:class:`pacman.model.constraints.abstract_constraint.AbstractConstraint`
        """
        pass



from setuptools import setup
from collections import defaultdict
import os

__version__ = None
exec(open("spinn_front_end_common/_version.py").read())
assert __version__

# Build a list of all project modules, as well as supplementary files
main_package = "spinn_front_end_common"
extensions = {".aplx", ".boot", ".cfg", ".json", ".sql", ".template", ".xml",
              ".xsd"}
main_package_dir = os.path.join(os.path.dirname(__file__), main_package)
start = len(main_package_dir)
packages = []
package_data = defaultdict(list)
for dirname, dirnames, filenames in os.walk(main_package_dir):
    if '__init__.py' in filenames:
        package = "{}{}".format(
            main_package, dirname[start:].replace(os.sep, '.'))
        packages.append(package)
    for filename in filenames:
        _, ext = os.path.splitext(filename)
        if ext in extensions:
            package = "{}{}".format(
                main_package, dirname[start:].replace(os.sep, '.'))
            package_data[package].append(filename)

setup(
    name="SpiNNFrontEndCommon",
    version=__version__,
    description="Common Spinnaker Front end functions",
    url="https://github.com/SpiNNakerManchester/SpiNNFrontEndCommon",
    packages=packages,
    package_data=package_data,
    install_requires=['SpiNNUtilities >= 1!4.0.0, < 1!5.0.0',
                      'SpiNNStorageHandlers >= 1!4.0.0, < 1!5.0.0',
                      'SpiNNMachine >= 1!4.0.0, < 1!5.0.0',
                      'SpiNNMan >= 1!4.0.0, < 1!5.0.0',
                      'SpiNNaker_PACMAN >= 1!4.0.0, < 1!5.0.0',
                      'SpiNNaker_DataSpecification >= 1!4.0.0, < 1!5.0.0',
                      'spalloc >= 1.0.0, < 2.0.0',
                      'requests >= 2.4.1',
                      'scipy >= 0.16.0',
                      'numpy',
                      'six']
)



from spinn_utilities.progress_bar import ProgressBar
from data_specification.constants import MAX_MEM_REGIONS
from spinn_front_end_common.utilities.exceptions import ConfigurationException

import logging
import os
import struct

logger = logging.getLogger(__name__)

MEM_MAP_SUBDIR_NAME = "memory_map_reports"


class MemoryMapOnHostChipReport(object):
    """ Report on memory usage
    """

    def __call__(
            self, report_default_directory, dsg_targets, transceiver,
            loaded_app_data_token):
        """ creates a report that states where in sdram each region is
        (read from machine)

        :param report_default_directory: the folder where reports are written
        :param dsg_targets: the map between placement and file writer
        :param transceiver: the spinnMan instance
        :param loaded_app_data_token: flag that app data has been loaded
        :rtype: None
        """

        if not loaded_app_data_token:
            raise ConfigurationException(
                "Needs to have loaded app data for this to work.")

        directory_name = os.path.join(
            report_default_directory, MEM_MAP_SUBDIR_NAME)
        if not os.path.exists(directory_name):
            os.makedirs(directory_name)

        progress = ProgressBar(dsg_targets, "Writing memory map reports")
        for (x, y, p) in progress.over(dsg_targets):
            file_name = os.path.join(
                directory_name,
                "memory_map_from_processor"
                "_{0:d}_{1:d}_{2:d}.txt".format(x, y, p))
            try:
                with open(file_name, "w") as f:
                    f.write("On chip data specification executor\n\n")

                    dsg_app_pointer_table_address_pointer = transceiver.\
                        get_user_0_register_address_from_core(x, y, p)

                    dsg_app_pointer_table_address = \
                        self._get_app_pointer_table(
                            transceiver, x, y,
                            dsg_app_pointer_table_address_pointer)

                    report_bytes = 4 * MAX_MEM_REGIONS

                    mem_map_report_data = buffer(transceiver.read_memory(
                        x, y, dsg_app_pointer_table_address, report_bytes))

                    offset = 0
                    for i in xrange(MAX_MEM_REGIONS):
                        region_address = int(struct.unpack_from(
                            "<I", mem_map_report_data, offset)[0])
                        offset += 4
                        f.write("Region {0:d}:\n\t start address: 0x{1:x}\n\t"
                                .format(i, region_address))
            except IOError:
                logger.error("Generate_placement_reports: Can't open file"
                             " {} for writing.".format(file_name))

    def _get_app_pointer_table(self, txrx, x, y, table_pointer):
        encoded_address = buffer(txrx.read_memory(x, y, table_pointer, 4))
        return int(struct.unpack_from("<I", encoded_address)[0]) + 8



# pacman imports
from pacman.model.decorators import overrides
from pacman.model.graphs.application import ApplicationEdge
from pacman.model.constraints.key_allocator_constraints \
    import FixedKeyAndMaskConstraint
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.resources import ResourceContainer, SDRAMResource
from pacman.model.routing_info import BaseKeyAndMask
from pacman.executor.injection_decorator import inject_items

# spinn front end common imports
from spinn_front_end_common.abstract_models import \
    AbstractProvidesOutgoingPartitionConstraints, AbstractHasAssociatedBinary
from spinn_front_end_common.abstract_models import \
    AbstractVertexWithEdgeToDependentVertices
from spinn_front_end_common.abstract_models \
    import AbstractGeneratesDataSpecification
from spinn_front_end_common.utilities import constants
from .command_sender_machine_vertex import CommandSenderMachineVertex
from spinn_front_end_common.utilities.utility_objs import ExecutableStartType


class CommandSender(
        ApplicationVertex, AbstractGeneratesDataSpecification,
        AbstractHasAssociatedBinary,
        AbstractProvidesOutgoingPartitionConstraints):
    """ A utility for sending commands to a vertex (possibly an external\
        device) at fixed times in the simulation
    """

    # all commands will use this mask
    _DEFAULT_COMMAND_MASK = 0xFFFFFFFF

    def __init__(self, label, constraints):

        ApplicationVertex.__init__(self, label, constraints, 1)

        self._timed_commands = list()
        self._commands_at_start_resume = list()
        self._commands_at_pause_stop = list()
        self._partition_id_to_keys = dict()
        self._keys_to_partition_id = dict()
        self._edge_partition_id_counter = 0
        self._vertex_to_key_map = dict()

    def add_commands(
            self, start_resume_commands, pause_stop_commands,
            timed_commands, vertex_to_send_to):
        """ Add commands to be sent down a given edge

        :param start_resume_commands: The commands to send when the simulation\
                starts or resumes from pause
        :type start_resume_commands: iterable of\
                :py:class:`spinn_front_end_common.utility_models.multi_cast_command.MultiCastCommand`
        :param pause_stop_commands: the commands to send when the simulation\
                stops or pauses after running
        :type pause_stop_commands: iterable of\
                :py:class:`spinn_front_end_common.utility_models.multi_cast_command.MultiCastCommand`
        :param timed_commands: The commands to send at specific times
        :type timed_commands: iterable of\
                :py:class:`spinn_front_end_common.utility_models.multi_cast_command.MultiCastCommand`
        :param vertex_to_send_to: The vertex these commands are to be sent to
        """

        # container for keys for partition mapping (remove duplicates)
        command_keys = set()
        self._vertex_to_key_map[vertex_to_send_to] = set()

        # update holders
        self._commands_at_start_resume.extend(start_resume_commands)
        self._commands_at_pause_stop.extend(pause_stop_commands)
        self._timed_commands.extend(timed_commands)

        for commands in (
                start_resume_commands, pause_stop_commands, timed_commands):
            for command in commands:
                # track keys
                command_keys.add(command.key)
                self._vertex_to_key_map[vertex_to_send_to].add(command.key)

        # create mapping between keys and partitions via partition constraint
        for key in command_keys:

            partition_id = "COMMANDS{}".format(self._edge_partition_id_counter)
            self._keys_to_partition_id[key] = partition_id
            self._partition_id_to_keys[partition_id] = key
            self._edge_partition_id_counter += 1

    @inject_items({
        "machine_time_step": "MachineTimeStep",
        "time_scale_factor": "TimeScaleFactor",
        "n_machine_time_steps": "RunTimeMachineTimeSteps"
    })
    @overrides(
        AbstractGeneratesDataSpecification.generate_data_specification,
        additional_arguments={
            "machine_time_step", "time_scale_factor", "n_machine_time_steps"
        })
    def generate_data_specification(
            self, spec, placement, machine_time_step, time_scale_factor,
            n_machine_time_steps):
        placement.vertex.generate_data_specification(
            spec, placement, machine_time_step, time_scale_factor,
            n_machine_time_steps)

    @overrides(ApplicationVertex.create_machine_vertex)
    def create_machine_vertex(
            self, vertex_slice, resources_required, label=None,
            constraints=None):
        return CommandSenderMachineVertex(
            constraints, resources_required, label,
            self._commands_at_start_resume, self._commands_at_pause_stop,
            self._timed_commands)

    @overrides(ApplicationVertex.get_resources_used_by_atoms)
    def get_resources_used_by_atoms(self, vertex_slice):

        sdram = (
            CommandSenderMachineVertex.get_timed_commands_bytes(
                self._timed_commands) +
            CommandSenderMachineVertex.get_n_command_bytes(
                self._commands_at_start_resume) +
            CommandSenderMachineVertex.get_n_command_bytes(
                self._commands_at_pause_stop) +
            constants.SYSTEM_BYTES_REQUIREMENT +
            CommandSenderMachineVertex.get_provenance_data_size(0) +
            (CommandSenderMachineVertex.get_number_of_mallocs_used_by_dsg() *
             constants.SARK_PER_MALLOC_SDRAM_USAGE))

        # Return the SDRAM and 1 core
        return ResourceContainer(sdram=SDRAMResource(sdram))

    @property
    @overrides(ApplicationVertex.n_atoms)
    def n_atoms(self):
        return 1

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        """ Return a string representation of the models binary

        """
        return CommandSenderMachineVertex.BINARY_FILE_NAME

    @overrides(AbstractVertexWithEdgeToDependentVertices.dependent_vertices)
    def dependent_vertices(self):
        return self._vertex_to_key_map.keys()

    def edges_and_partitions(self):
        edges = list()
        partition_ids = list()
        keys_added = set()
        for vertex in self._vertex_to_key_map:
            for key in self._vertex_to_key_map[vertex]:
                if key not in keys_added:
                    keys_added.add(key)
                    app_edge = ApplicationEdge(self, vertex)
                    edges.append(app_edge)
                    partition_ids.append(self._keys_to_partition_id[key])
        return edges, partition_ids

    @overrides(AbstractProvidesOutgoingPartitionConstraints.
               get_outgoing_partition_constraints)
    def get_outgoing_partition_constraints(self, partition):
        return [FixedKeyAndMaskConstraint([
            BaseKeyAndMask(
                self._partition_id_to_keys[partition.identifier],
                self._DEFAULT_COMMAND_MASK)
        ])]

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableStartType.USES_SIMULATION_INTERFACE



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractRecordable(object):
    """ Indicates that an object might record some data in to SDRAM
    """
    __slots__ = ()

    @abstractmethod
    def is_recording(self):
        """ Deduce if the recorder is actually recording
        """



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase
from spinn_utilities.abstract_base import abstractproperty
from spinn_utilities.abstract_base import abstractmethod


@add_metaclass(AbstractBase)
class SimulatorInterface(object):

    __slots__ = ()

    @abstractmethod
    def add_socket_address(self, x):
        pass

    @abstractproperty
    def buffer_manager(self):
        pass

    @abstractproperty
    def config(self):
        pass

    @abstractproperty
    def graph_mapper(self):
        pass

    @abstractproperty
    def has_ran(self):
        pass

    @abstractmethod
    def verify_not_running(self):
        pass

    @abstractproperty
    def increment_none_labelled_vertex_count(self):
        pass

    @abstractproperty
    def machine(self):
        pass

    @abstractproperty
    def machine_time_step(self):
        pass

    @abstractproperty
    def no_machine_time_steps(self):
        pass

    @abstractproperty
    def none_labelled_vertex_count(self):
        pass

    @abstractproperty
    def placements(self):
        pass

    @abstractproperty
    def run(self, run_time):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractproperty
    def transceiver(self):
        pass

    @abstractproperty
    def use_virtual_board(self):
        pass



from spinn_utilities.progress_bar import ProgressBar

from spinn_front_end_common.utilities import exceptions
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary
from spinnman.model import ExecutableTargets


class GraphBinaryGatherer(object):
    """ Extracts binaries to be executed
    """

    __slots__ = []

    def __call__(
            self, placements, graph, executable_finder, graph_mapper=None):
        executable_targets = ExecutableTargets()
        progress = ProgressBar(graph.n_vertices, "Finding binaries")
        for vertex in progress.over(graph.vertices):
            placement = placements.get_placement_of_vertex(vertex)
            self._get_binary(
                placement, vertex, executable_targets, executable_finder)

            if graph_mapper is not None:
                associated_vertex = graph_mapper.get_application_vertex(vertex)
                self._get_binary(
                    placement, associated_vertex, executable_targets,
                    executable_finder)

        return executable_targets

    @staticmethod
    def _get_binary(
            placement, associated_vertex, executable_targets,
            executable_finder):
        # if the vertex cannot generate a DSG, ignore it
        if not isinstance(associated_vertex, AbstractHasAssociatedBinary):
            return None

        # Get name of binary from vertex
        binary_name = associated_vertex.get_binary_file_name()

        # Attempt to find this within search paths
        binary_path = executable_finder.get_executable_path(binary_name)
        if binary_path is None:
            raise exceptions.ExecutableNotFoundException(binary_name)

        executable_targets.add_processor(
            binary_path, placement.x, placement.y, placement.p)



from threading import Thread
from collections import OrderedDict

from spinn_front_end_common.utilities.database import DatabaseConnection

from spinnman.utilities.utility_functions import send_port_trigger_message
from spinnman.messages.eieio.data_messages import EIEIODataMessage
from spinnman.messages.eieio import EIEIOType
from spinnman.connections import ConnectionListener
from spinnman.connections.udp_packet_connections import EIEIOConnection
from spinnman.messages.eieio.data_messages import KeyPayloadDataElement

import logging

logger = logging.getLogger(__name__)


# The maximum number of 32-bit keys that will fit in a packet
_MAX_FULL_KEYS_PER_PACKET = 63

# The maximum number of 16-bit keys that will fit in a packet
_MAX_HALF_KEYS_PER_PACKET = 127


class LiveEventConnection(DatabaseConnection):
    """ A connection for receiving and sending live events from and to\
        SpiNNaker
    """

    def __init__(self, live_packet_gather_label, receive_labels=None,
                 send_labels=None, local_host=None, local_port=19999,
                 machine_vertices=False):
        """

        :param live_packet_gather_label: The label of the LivePacketGather\
                    vertex to which received events are being sent
        :param receive_labels: Labels of vertices from which live events\
                    will be received.
        :type receive_labels: iterable of str
        :param send_labels: Labels of vertices to which live events will be\
                    sent
        :type send_labels: iterable of str
        :param local_host: Optional specification of the local hostname or\
                    ip address of the interface to listen on
        :type local_host: str
        :param local_port: Optional specification of the local port to listen\
                    on.  Must match the port that the toolchain will send the\
                    notification on (19999 by default)
        :type local_port: int

        """

        DatabaseConnection.__init__(
            self, self._start_resume_callback, self._stop_pause_callback,
            local_host=local_host, local_port=local_port)

        self.add_database_callback(self._read_database_callback)

        self._live_packet_gather_label = live_packet_gather_label
        self._receive_labels = receive_labels
        self._send_labels = send_labels
        self._machine_vertices = machine_vertices
        self._sender_connection = None
        self._send_address_details = dict()
        self._atom_id_to_key = dict()
        self._key_to_atom_id_and_label = dict()
        self._live_event_callbacks = list()
        self._start_resume_callbacks = dict()
        self._pause_stop_callbacks = dict()
        self._init_callbacks = dict()
        if receive_labels is not None:
            for label in receive_labels:
                self._live_event_callbacks.append(list())
                self._start_resume_callbacks[label] = list()
                self._pause_stop_callbacks[label] = list()
                self._init_callbacks[label] = list()
        if send_labels is not None:
            for label in send_labels:
                self._start_resume_callbacks[label] = list()
                self._pause_stop_callbacks[label] = list()
                self._init_callbacks[label] = list()
        self._receivers = dict()
        self._listeners = dict()

    def add_init_callback(self, label, init_callback):
        """ Add a callback to be called to initialise a vertex

        :param label: The label of the vertex to be notified about. Must\
                    be one of the vertices listed in the constructor
        :type label: str
        :param init_callback: A function to be called to initialise the\
                    vertex.  This should take as parameters the label of the\
                    vertex, the number of neurons in the population,\
                    the run time of the simulation in milliseconds, and the\
                    simulation timestep in milliseconds
        :type init_callback: function(str, int, float, float) -> None
        """
        self._init_callbacks[label].append(init_callback)

    def add_receive_callback(self, label, live_event_callback):
        """ Add a callback for the reception of live events from a vertex

        :param label: The label of the vertex to be notified about. Must\
                    be one of the vertices listed in the constructor
        :type label: str
        :param live_event_callback: A function to be called when events\
                    are received.  This should take as parameters the label\
                    of the vertex, the simulation timestep when the event\
                    occurred, and an array-like of atom ids.
        :type live_event_callback: function(str, int, [int]) -> None
        """
        label_id = self._receive_labels.index(label)
        self._live_event_callbacks[label_id].append(live_event_callback)

    def add_start_callback(self, label, start_callback):
        """ Add a callback for the start of the simulation

        :param start_callback: A function to be called when the start\
                    message has been received.  This function should take the\
                    label of the referenced vertex, and an instance of\
                    this class, which can be used to send events
        :type start_callback: function(str, \
                    :py:class:`SpynnakerLiveEventConnection`) -> None
        :param label: the label of the function to be sent
        :type label: str
        """
        logger.warning(
            "the method \"add_start_callback(label, start_callback)\" is in "
            "deprecation, and will be replaced with the method "
            "\"add_start_resume_callback(label, start_resume_callback)\" in a "
            "future release.")
        self.add_start_resume_callback(label, start_callback)

    def add_start_resume_callback(self, label, start_resume_callback):
        self._start_resume_callbacks[label].append(start_resume_callback)

    def add_pause_stop_callback(self, label, pause_stop_callback):
        """ Add a callback for the pause and stop state of the simulation

        :param label: the label of the function to be sent
        :type label: str
        :param pause_stop_callback: A function to be called when the pause\
                    or stop message has been received.
                    This function should take the\
                    label of the referenced vertex, and an instance of\
                    this class, which can be used to send events
        :type pause_stop_callback: function(str, \
                    :py:class:`SpynnakerLiveEventConnection`) -> None
        :rtype: None
        """
        self._pause_stop_callbacks[label].append(pause_stop_callback)

    def _read_database_callback(self, database_reader):
        self._handle_possible_rerun_state()

        vertex_sizes = OrderedDict()
        run_time_ms = database_reader.get_configuration_parameter_value(
            "runtime")
        machine_timestep_ms = \
            database_reader.get_configuration_parameter_value(
                "machine_time_step") / 1000.0

        if self._send_labels is not None:
            self._sender_connection = EIEIOConnection()
            for send_label in self._send_labels:
                ip_address, port = None, None
                if self._machine_vertices:
                    ip_address, port = \
                        database_reader.get_machine_live_input_details(
                            send_label)
                else:
                    ip_address, port = database_reader.get_live_input_details(
                        send_label)
                self._send_address_details[send_label] = (ip_address, port)
                if self._machine_vertices:
                    key, _ = database_reader.get_machine_live_input_key(
                        send_label)
                    self._atom_id_to_key[send_label] = {0: key}
                    vertex_sizes[send_label] = 1
                else:
                    self._atom_id_to_key[send_label] = \
                        database_reader.get_atom_id_to_key_mapping(send_label)
                    vertex_sizes[send_label] = len(
                        self._atom_id_to_key[send_label])

        if self._receive_labels is not None:

            label_id = 0
            for receive_label in self._receive_labels:
                host, port, strip_sdp = None, None, None
                if self._machine_vertices:
                    host, port, strip_sdp, board_address = \
                        database_reader.get_machine_live_output_details(
                            receive_label, self._live_packet_gather_label)
                else:
                    host, port, strip_sdp, board_address = \
                        database_reader.get_live_output_details(
                            receive_label, self._live_packet_gather_label)
                if not strip_sdp:
                    raise Exception("Currently, only ip tags which strip the"
                                    " SDP headers are supported")
                if port not in self._receivers:
                    receiver = EIEIOConnection(local_port=port)
                    listener = ConnectionListener(receiver)
                    listener.add_callback(self._receive_packet_callback)
                    listener.start()
                    self._receivers[port] = receiver
                    self._listeners[port] = listener

                send_port_trigger_message(receiver, board_address)
                logger.info(
                    "Listening for traffic from {} on {}:{}".format(
                        receive_label, host, port))

                if self._machine_vertices:
                    key, _ = database_reader.get_machine_live_output_key(
                        receive_label, self._live_packet_gather_label)
                    self._key_to_atom_id_and_label[key] = (0, label_id)
                    vertex_sizes[receive_label] = 1
                else:
                    key_to_atom_id = \
                        database_reader.get_key_to_atom_id_mapping(
                            receive_label)
                    for (key, atom_id) in key_to_atom_id.iteritems():
                        self._key_to_atom_id_and_label[key] = (
                            atom_id, label_id)
                    vertex_sizes[receive_label] = len(key_to_atom_id)

                label_id += 1

        for (label, vertex_size) in vertex_sizes.iteritems():
            for init_callback in self._init_callbacks[label]:
                init_callback(
                    label, vertex_size, run_time_ms, machine_timestep_ms)

    def _handle_possible_rerun_state(self):
        # reset from possible previous calls
        if self._sender_connection is not None:
            self._sender_connection.close()
            self._sender_connection = None
        for port in self._receivers:
            self._receivers[port].close()
        self._receivers = dict()
        for port in self._listeners:
            self._listeners[port].close()
        self._listeners = dict()

    def _start_resume_callback(self):
        for (label, callbacks) in self._start_resume_callbacks.iteritems():
            for callback in callbacks:
                callback_thread = Thread(
                    target=callback, args=(label, self),
                    verbose=True,
                    name="start_resume callback thread for "
                         "live_event_connection"
                         "{}:{}".format(
                             self._local_port, self._local_ip_address))
                callback_thread.start()

    def _stop_pause_callback(self):
        for (label, callbacks) in self._pause_stop_callbacks.iteritems():
            for callback in callbacks:
                callback_thread = Thread(
                    target=callback, args=(label, self),
                    verbose=True,
                    name="pause_stop callback thread for live_event_connection"
                         "{}:{}".format(
                             self._local_port, self._local_ip_address))
                callback_thread.start()

    def _receive_packet_callback(self, packet):
        try:
            header = packet.eieio_header
            if header.is_time:
                key_times_labels = OrderedDict()
                while packet.is_next_element:
                    element = packet.next_element
                    time = element.payload
                    key = element.key
                    if key in self._key_to_atom_id_and_label:
                        (atom_id, label_id) = self._key_to_atom_id_and_label[
                            key]
                        if time not in key_times_labels:
                            key_times_labels[time] = dict()
                        if label_id not in key_times_labels[time]:
                            key_times_labels[time][label_id] = list()
                        key_times_labels[time][label_id].append(atom_id)

                for time in key_times_labels.iterkeys():
                    for label_id in key_times_labels[time].iterkeys():
                        label = self._receive_labels[label_id]
                        for callback in self._live_event_callbacks[label_id]:
                            callback(
                                label, time, key_times_labels[time][label_id])
            else:
                while packet.is_next_element:
                    element = packet.next_element
                    key = element.key
                    if key in self._key_to_atom_id_and_label:
                        (atom_id, label_id) = self._key_to_atom_id_and_label[
                            key]
                        for callback in self._live_event_callbacks[label_id]:
                            if isinstance(element, KeyPayloadDataElement):
                                callback(
                                    self._receive_labels[label_id], atom_id,
                                    element.payload)
                            else:
                                callback(
                                    self._receive_labels[label_id], atom_id)
        except:
            logger.warn("problem handling received packet", exc_info=True)

    def send_event(self, label, atom_id, send_full_keys=False):
        """ Send an event from a single atom

        :param label: The label of the vertex from which the event will\
                    originate
        :type label: str
        :param atom_id: The id of the atom sending the event
        :type atom_id: int
        :param send_full_keys: Determines whether to send full 32-bit keys,\
                    getting the key for each atom from the database, or\
                    whether to send 16-bit atom ids directly
        :type send_full_keys: bool
        """
        self.send_events(label, [atom_id], send_full_keys)

    def send_events(self, label, atom_ids, send_full_keys=False):
        """ Send a number of events

        :param label: The label of the vertex from which the events will\
                    originate
        :type label: str
        :param atom_ids: array-like of atom ids sending events
        :type atom_ids: [int]
        :param send_full_keys: Determines whether to send full 32-bit keys,\
                    getting the key for each atom from the database, or\
                    whether to send 16-bit atom ids directly
        :type send_full_keys: bool
        """
        max_keys = _MAX_HALF_KEYS_PER_PACKET
        if send_full_keys:
            max_keys = _MAX_FULL_KEYS_PER_PACKET

        pos = 0
        while pos < len(atom_ids):

            if send_full_keys:
                message = EIEIODataMessage.create(EIEIOType.KEY_32_BIT)
            else:
                message = EIEIODataMessage.create(EIEIOType.KEY_16_BIT)

            events_in_packet = 0
            while pos < len(atom_ids) and events_in_packet < max_keys:
                key = atom_ids[pos]
                if send_full_keys:
                    key = self._atom_id_to_key[label][key]
                message.add_key(key)
                pos += 1
                events_in_packet += 1
            ip_address, port = self._send_address_details[label]
            self._sender_connection.send_eieio_message_to(
                message, ip_address, port)

    def close(self):
        DatabaseConnection.close(self)



from collections import defaultdict

from spinn_utilities.progress_bar import ProgressBar

from data_specification.utility_calls \
    import get_data_spec_and_file_writer_filename

from spinn_front_end_common.abstract_models import \
    AbstractGeneratesDataSpecification
from spinn_front_end_common.utilities.exceptions import ConfigurationException


class GraphDataSpecificationWriter(object):
    """ Executes data specification generation
    """

    __slots__ = (
        # Dict of sdram usage by chip coordinates
        "_sdram_usage",
        # Dict of list of region sizes by vertex
        "_region_sizes",
        # Dict of list of vertices by chip coordinates
        "_vertices_by_chip"
    )

    def __init__(self):
        self._sdram_usage = defaultdict(lambda: 0)
        self._region_sizes = dict()
        self._vertices_by_chip = defaultdict(list)

    def __call__(
            self, placements, hostname,
            report_default_directory, write_text_specs,
            app_data_runtime_folder, machine, graph_mapper=None):
        """

        :param placements: placements of machine graph to cores
        :param hostname: spinnaker machine name
        :param report_default_directory: the location where reports are stored
        :param write_text_specs:\
            True if the textual version of the specification is to be written
        :param app_data_runtime_folder:\
            Folder where data specifications should be written to
        :param machine: the python representation of the spinnaker machine
        :param graph_mapper:\
            the mapping between application and machine graph

        :return: dsg targets (map of placement tuple and filename)
        """

        # iterate though vertices and call generate_data_spec for each
        # vertex
        dsg_targets = dict()

        progress = ProgressBar(
            placements.n_placements, "Generating data specifications")
        for placement in progress.over(placements.placements):
            # Try to generate the data spec for the placement
            generated = self._generate_data_spec_for_vertices(
                placement, placement.vertex, dsg_targets, hostname,
                report_default_directory, write_text_specs,
                app_data_runtime_folder, machine)

            # If the spec wasn't generated directly, and there is an
            # application vertex, try with that
            if not generated and graph_mapper is not None:
                associated_vertex = graph_mapper.get_application_vertex(
                    placement.vertex)
                self._generate_data_spec_for_vertices(
                    placement, associated_vertex, dsg_targets, hostname,
                    report_default_directory, write_text_specs,
                    app_data_runtime_folder, machine)

        return dsg_targets

    def _generate_data_spec_for_vertices(
            self, placement, vertex, dsg_targets, hostname,
            report_default_directory, write_text_specs,
            app_data_runtime_folder, machine):
        """

        :param placement: placement of machine graph to cores
        :param vertex: the specific vertex to write dsg for.
        :param hostname: spinnaker machine name
        :param report_default_directory: the location where reports are stored
        :param write_text_specs:\
            True if the textual version of the specification is to be written
        :param app_data_runtime_folder: \
            Folder where data specifications should be written to
        :param machine: the python representation of the spinnaker machine
        :return: True if the vertex was data specable, False otherwise
        """

        # if the vertex can generate a DSG, call it
        if not isinstance(vertex, AbstractGeneratesDataSpecification):
            return False

        # build the writers for the reports and data
        data_writer_filename, spec = get_data_spec_and_file_writer_filename(
            placement.x, placement.y, placement.p, hostname,
            report_default_directory,
            write_text_specs, app_data_runtime_folder)

        # link dsg file to vertex
        dsg_targets[placement.x, placement.y, placement.p] = \
            data_writer_filename

        # generate the dsg file
        vertex.generate_data_specification(spec, placement)

        # Check the memory usage
        self._region_sizes[placement.vertex] = spec.region_sizes
        self._vertices_by_chip[placement.x, placement.y].append(
            placement.vertex)
        self._sdram_usage[placement.x, placement.y] += sum(
            spec.region_sizes)
        if (self._sdram_usage[placement.x, placement.y] <=
                machine.get_chip_at(placement.x, placement.y).sdram.size):
            return True

        # creating the error message which contains the memory usage of
        #  what each core within the chip uses and its original
        # estimate.
        memory_usage = "\n".join((
            "    {}: {} (total={}, estimated={})".format(
                vert, self._region_sizes[vert],
                sum(self._region_sizes[vert]),
                vert.resources_required.sdram.get_value())
            for vert in self._vertices_by_chip[placement.x, placement.y]))

        raise ConfigurationException(
            "Too much SDRAM has been used on {}, {}.  Vertices and"
            " their usage on that chip is as follows:\n{}".format(
                placement.x, placement.y, memory_usage))



from spinn_utilities.progress_bar import ProgressBar

# pacman imports
from pacman.utilities.utility_objs import ResourceTracker
from pacman.utilities.algorithm_utilities import placer_algorithm_utilities

# general imports
import logging
logger = logging.getLogger(__name__)


class GraphMeasurer(object):
    """ Works out how many chips a machine graph needs
    """

    __slots__ = []

    def __call__(self, machine_graph, machine):
        """

        :param machine_graph: The machine_graph to measure
        :type machine_graph:\
                    :py:class:`pacman.model.graph.machine.MachineGraph`
        :return: The size of the graph in number of chips
        :rtype: int
        """

        # check that the algorithm can handle the constraints
        ResourceTracker.check_constraints(machine_graph.vertices)

        ordered_vertices = \
            placer_algorithm_utilities.sort_vertices_by_known_constraints(
                machine_graph.vertices)

        # Iterate over vertices and allocate
        progress = ProgressBar(machine_graph.n_vertices, "Measuring the graph")
        resource_tracker = ResourceTracker(machine)
        for vertex in progress.over(ordered_vertices):
            resource_tracker.allocate_constrained_resources(
                vertex.resources_required, vertex.constraints)
        return len(resource_tracker.keys)



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractVertexWithEdgeToDependentVertices(object):
    """ A vertex with a dependent vertices, which should be connected to this\
        vertex by an edge directly to each of them
    """

    __slots__ = ()

    @abstractmethod
    def dependent_vertices(self):
        """ Return the vertices which this vertex depends upon
        """

    @abstractmethod
    def edge_partition_identifiers_for_dependent_vertex(self, vertex):
        """ Return the dependent edge identifiers for this vertex
        """
